import numpy as np

from ...constants import log
from ...constants import tol_path as tol
from ..entities   import Line, Arc, BSpline
from ..util       import angles_to_threepoint, is_ccw
from ...util      import is_binary_file, multi_dict

from collections import deque

# unit codes
_DXF_UNITS = {1  : 'inches',
              2  : 'feet',
              3  : 'miles',
              4  : 'millimeters', 
              5  : 'centimeters',
              6  : 'meters',
              7  : 'kilometers',
              8  : 'microinches', 
              9  : 'mils',
              10 : 'yards',
              11 : 'angstroms',
              12 : 'nanometers', 
              13 : 'microns',
              14 : 'decimeters', 
              15 : 'decameters', 
              16 : 'hectometers', 
              17 : 'gigameters',
              18 : 'AU',
              19 : 'light years', 
              20 : 'parsecs'}

def get_key(blob, field, code):
    try: 
        line = blob[np.nonzero(blob[:,1] == field)[0][0]+1]
    except IndexError: 
        return None
    if line[0] == code:
        return int(line[1])
    else: 
        return None
    
def load_dxf(file_obj):
    '''
    Load a DXF file to a dictionary containing vertices and entities.

    Arguments
    ----------
    file_obj: file or file- like object (has object.read method)

    Returns
    ----------
    result: dict, keys are  entities, vertices and metadata
    '''
    def convert_line(e_data):
        e = dict(e_data)
        entities.append(Line(len(vertices) + np.arange(2)))
        vertices.extend(np.array([[e['10'], e['20']],
                                  [e['11'], e['21']]], dtype=np.float))        
    def convert_circle(e_data):
        e = dict(e_data)
        R = float(e['40'])
        C = np.array([e['10'], e['20']]).astype(float)
        points = angles_to_threepoint([0, np.pi], C[0:2], R)
        entities.append(Arc(points=(len(vertices) + np.arange(3)), closed=True))
        vertices.extend(points)

    def convert_arc(e_data):
        e = dict(e_data)
        R = float(e['40'])
        C = np.array([e['10'], e['20']], dtype=np.float)
        A = np.radians(np.array([e['50'], e['51']], dtype=np.float))
        points = angles_to_threepoint(A, C[0:2], R)
        entities.append(Arc(len(vertices) + np.arange(3), closed=False))
        vertices.extend(points)
    def convert_polyline(e_data):
        e     = multi_dict(e_data)
        lines = np.column_stack((e['10'], e['20'])).astype(np.float)
        entities.append(Line(np.arange(len(lines))+len(vertices)))
        vertices.extend(lines)

    def convert_bspline(e_data):
        # in the DXF there are n points and n ordered fields 
        # with the same group code 
        e      = multi_dict(e_data)
        points = np.column_stack((e['10'], e['20'])).astype(np.float)
        knots  = np.array(e['40']).astype(float)
        # check euclidean distance to see if closed
        closed = np.linalg.norm(points[0] - points[-1]) < tol.merge
        # if it is closed, make sure it is CCW for later polygon happiness
        if closed and (not is_ccw(np.vstack((points, points[0])))):
            points = points[::-1]
        entities.append(BSpline(points = np.arange(len(points))+len(vertices),
                                knots  = knots,
                                closed = closed))
        vertices.extend(points)

    if is_binary_file(file_obj): 
        raise ValueError("Binary DXF is unsupported!")
    
    # in a DXF file, lines come in pairs, 
    # a group code then the next line is the value
    # we are removing all whitespace then splitting with the
    # splitlines function which uses the universal newline method
    raw  = str(file_obj.read().decode('ascii').upper().replace(' ', ''))
    # if this reshape fails, it means the DXF is malformed
    blob = np.array(str.splitlines(raw)).reshape((-1,2))
    
    # get the section which contains the header in the DXF file
    endsec       = np.nonzero(blob[:,1] == 'ENDSEC')[0]
    header_start = np.nonzero(blob[:,1] == 'HEADER')[0][0]
    header_end   = endsec[np.searchsorted(endsec, header_start)]
    header_blob = blob[header_start:header_end]

    # get the section which contains entities in the DXF file
    entity_start = np.nonzero(blob[:,1] == 'ENTITIES')[0][0]
    entity_end   = endsec[np.searchsorted(endsec, entity_start)]
    entity_blob  = blob[entity_start:entity_end]
    
    # store metadata pulled from the header of the DXF
    metadata = dict()
    units    = get_key(header_blob, '$INSUNITS', '70')
    if units in _DXF_UNITS:
        metadata['units'] = _DXF_UNITS[units]
    else: 
        log.warning('DXF doesn\'t have units specified!')

    # find the start points of entities
    #group_check = np.logical_or(entity_blob[:,0] == '0', 
    #                            entity_blob[:,0] == '5')
    group_check = entity_blob[:,0] == '0'
    inflection = np.nonzero(group_check)[0]

    #inflection = np.nonzero(np.logical_and(group_check[:-1], 
    #                                       group_check[:-1] == group_check[1:]))[0]
    loaders = {'LINE'       : convert_line,
               'LWPOLYLINE' : convert_polyline,
               'ARC'        : convert_arc,
               'CIRCLE'     : convert_circle,
               'SPLINE'     : convert_bspline}
    vertices = deque()
    entities = deque()
    
    for chunk in np.array_split(entity_blob, inflection):
        if len(chunk) > 2:
            entity_type = chunk[0][1]
            if entity_type in loaders:
                loaders[entity_type](chunk)
            else:
                log.debug('Entity type %s not supported', entity_type)
            
    result = {'vertices' : np.vstack(vertices).astype(np.float64),
              'entities' : np.array(entities),
              'metadata' : metadata}
              
    return result
            



import numpy as np

from ..constants import res_path as res
from ..constants import tol_path as tol

def discretize_bezier(points, count=None, scale=1.0):
    '''
    Arguments
    ----------
    points:  (o,d) list of points of the bezier. The first and last 
             points should be the start and end of the curve.  
             For a 2D cubic bezier, order o=3, dimension d=2

    Returns
    ----------
    discrete: (n,d) list of points, a polyline representation
              of the bezier curve which respects constants.RES_LENGTH
    '''
    def compute(t):
        # compute discrete points given a sampling t
        t_d      = 1.0 - t
        n        = len(points) - 1
        # binomial coefficents, i, and each point
        iterable = zip(binomial(n), np.arange(n+1), points)
        stacked  = [((t**i)*(t_d**(n-i))).reshape((-1,1))*p*c for c,i,p in iterable]
        discrete = np.sum(stacked, axis=0)
        return discrete
    
    # make sure we have a numpy array
    points  = np.array(points)

    if count is None:
        # how much distance does a small percentage of the curve take
        # this is so we can figure out how finely we have to sample t   
        norm  = np.linalg.norm(np.diff(points, axis=0), axis=1).sum()
        count = np.ceil(norm / (res.seg_frac * scale))
        count  = int(np.clip(count, 
                             res.min_sections*len(points), 
                             res.max_sections*len(points)))
    result = compute(np.linspace(0.0, 1.0, count))

    test =  np.sum((result[[0,-1]] - points[[0,-1]])**2, axis=1)
    assert (test < tol.merge).all()
    assert len(result) >= 2

    return result

def discretize_bspline(control, knots, count=None, scale=1.0):
    '''
    Given a B-Splines control points and knot vector, return
    a sampled version of the curve.

    Arguments
    ----------
    control:  (o,d) list of control points of the b- spline. 
    knots:    (j) list of knots 
    count:    int, number of sections to discretize the spline in to.
              If not specified, RES_LENGTH will be used to inform this. 

    Returns
    ----------
    discrete: (count,d) list of points, a polyline of the B-spline.
    '''

    # evaluate the b-spline using scipy/fitpack
    from scipy.interpolate import splev
    # (n, d) control points where d is the dimension of vertices
    control = np.array(control)
    degree  = len(knots) - len(control) - 1
    if count is None:
        norm  = np.linalg.norm(np.diff(control, axis=0), axis=1).sum()
        count = int(np.clip(norm / (res.seg_frac*scale),
                            res.min_sections*len(control), 
                            res.max_sections*len(control)))
                            
    ipl      = np.linspace(knots[0], knots[-1], count)
    discrete = splev(ipl, [knots, control.T, degree])
    discrete = np.column_stack(discrete)
    return discrete
        
def binomial(n):
    '''
    Return all binomial coefficents for a given order.
    
    For n > 5, scipy.special.binom is used, below we hardcode
    to avoid the scipy.special dependancy. 
    '''
    if   n == 1: return [1,1]
    elif n == 2: return [1,2,1]
    elif n == 3: return [1,3,3,1]
    elif n == 4: return [1,4,6,4,1]
    elif n == 5: return [1,5,10,10,5,1]
    else:
        from scipy.special import binom
        return binom(n,np.arange(n+1))



import numpy as np

from .util      import zero_pad
from .grouping  import group_rows
from .constants import log, _log_time


_MIN_BIN_COUNT = 20
_TOL_FREQ      = 1e-3
    
def rotationally_invariant_identifier(mesh, length=6):
    '''
    Given an input mesh, return a vector or string that has the following properties:
    * invariant to rotation of the mesh
    * robust to different tesselation of the surfaces
    * meshes that are similar but not identical return values that are close in euclidean distance

    Does this by computing the area- weighted distribution of the radius (from the center of mass).

    Arguments
    ---------
    mesh:    Trimesh
    length:  number of terms to compute of the identifier

    Returns
    ---------
    identifer: (length) float array of unique identifier
    '''

    frequency_count = int(length - 2)

    # calculate the mass properties of the mesh, which is doing a surface integral to
    # find the center of volume of the mesh
    mass_properties = mesh.mass_properties(skip_inertia=True)
    vertex_radii    = np.sum((mesh.vertices.view(np.ndarray) - mesh.center_mass)**2, 
                             axis=1) **.5
    
    # since we will be computing the shape distribution of the radii, we need to make sure there
    # are enough values to populate more than one sample per bin.  
    bin_count = int(np.min([256, 
                            mesh.vertices.shape[0] * 0.2, 
                            mesh.faces.shape[0]    * 0.2]))
    
    # if any of the frequency checks fail, we will use this zero length vector as the 
    # formatted information for the identifier
    freq_formatted = np.zeros(frequency_count)

    if bin_count > _MIN_BIN_COUNT:
        face_area   = mesh.area_faces
        face_radii  = vertex_radii[mesh.faces].reshape(-1)
        area_weight = np.tile((face_area.reshape((-1,1))*(1.0/3.0)), (1,3)).reshape(-1)

        if face_radii.std() > 1e-3:
            freq_formatted = fft_freq_histogram(face_radii, 
                                                bin_count=bin_count, 
                                                frequency_count=frequency_count,
                                                weight=area_weight)
        
    # using the volume (from surface integral), surface area, and top frequencies
    identifier = np.hstack((mass_properties['volume'],
                            mass_properties['surface_area'],
                            freq_formatted))
    return identifier

def fft_freq_histogram(data, bin_count, frequency_count=4, weight=None):
    data = np.reshape(data, -1)
    if weight is None: 
        weight = np.ones(len(data))

    hist, bin_edges = np.histogram(data,
                                   weights = weight,
                                   bins    = bin_count)
    # we calculate the fft of the radius distribution
    fft  = np.abs(np.fft.fft(hist))
    # the magnitude is dependant on our weighting being good
    # frequency should be more solid in more cases 
    freq = np.fft.fftfreq(data.size, d=(bin_edges[1] - bin_edges[0])) + bin_edges[0]

    # now we must select the top FREQ_COUNT frequencies
    # if there are a bunch of frequencies whose components are very close in magnitude,
    # just picking the top FREQ_COUNT of them is non-deterministic
    # thus we take the top frequencies which have a magnitude that is distingushable 
    # and we zero pad if this means fewer values available
    fft_top = fft.argsort()[-(frequency_count + 1):]
    fft_ok  = np.diff(fft[fft_top]) > _TOL_FREQ
    if fft_ok.any():
        fft_start = np.nonzero(fft_ok)[0][0] + 1 
        fft_top   = fft_top[fft_start:]
        freq_final = np.sort(freq[fft_top])
    else:
        freq_final = []

    freq_formatted = zero_pad(freq_final, frequency_count)
    return freq_formatted

@_log_time
def merge_duplicates(meshes):
    '''
    Given a list of meshes, find meshes which are duplicates and merge them.

    Arguments
    ---------
    meshes: (n) list of meshes

    Returns
    ---------
    merged: (m) list of meshes where (m <= n)
    '''
    # so we can use advanced indexing
    meshes = np.array(meshes)
    # by default an identifier is a 1D float array with 6 elements
    hashes = [i.identifier for i in meshes]
    groups = group_rows(hashes, digits=1)
    merged = [None] * len(groups)
    for i, group in enumerate(groups):
        quantity = 0
        metadata = {}
        for mesh in meshes[group]:
            # if metadata exists don't nuke it
            if 'quantity' in mesh.metadata:
                quantity += mesh.metadata['quantity']
            else: 
                quantity += 1
            metadata.update(mesh.metadata)
 
        metadata['quantity'] = int(quantity)
        metadata['original_index'] = group

        merged[i] = meshes[group[0]]
        merged[i].metadata = metadata 
    log.info('merge_duplicates reduced part count from %d to %d', 
             len(meshes),
             len(merged))
    return np.array(merged)



import numpy as np
from colorsys import hsv_to_rgb
from collections import deque

from . import util
from .util      import is_sequence, is_shape, Cache, DataStore
from .constants import log

COLORS = {'red'    : [205,59,34,255],
          'purple' : [150,111,214,255],
          'blue'   : [119,158,203,255],
          'brown'  : [160,85,45,255]}
COLOR_DTYPE = np.dtype(np.uint8)
DEFAULT_COLOR = np.array(COLORS['purple'], dtype=COLOR_DTYPE)

class VisualAttributes(object):
    '''
    Hold the visual attributes (usually colors) for a mesh. 

    This is a bit of a dumpster fire and probably needs a re-write
    '''
    def __init__(self, mesh=None, dtype=None, **kwargs):
        self.mesh = mesh

        self._validate = True
        self._data = DataStore()
        self._cache = Cache(id_function = self.md5)

        if dtype is None: 
            dtype = COLOR_DTYPE
        self.dtype = dtype

        colors = _kwargs_to_color(mesh, **kwargs)
        self.vertex_colors, self.face_colors = colors

    def choose(self):
        '''
        If both face and vertex colors are defined, choose one of them.
        '''
        if all(self._set.values()):
            sig_face   = self._data['face_colors'].ptp(axis=0).sum()
            sig_vertex = self._cache['vertex_colors'].ptp(axis=0).sum()
            if sig_face > sig_vertex:
                self.vertex_colors = None
            else:
                self.face_colors = None
        
    @property
    def _set(self):
        result = {'face'   : is_shape(self._data['face_colors'], (-1, (3,4))),
                  'vertex' : is_shape(self._cache['vertex_colors'], (-1,(3,4)))}
        return result

    @property
    def defined(self):
        defined = any(self._set.values()) 
        defined = defined and self.mesh is not None
        return defined

    @property
    def transparency(self):
        '''
        Returns
        ------------
        transparency: bool, does the visual attributes contain any transparency
        '''
        cached = self._cache.get('transparency')
        if cached is not None: 
            return cached
        transparency = False
        color_max = (2**(COLOR_DTYPE.itemsize*8)) - 1
        if self._set['face']:
            transparency = (is_shape(self._data['face_colors'], (-1,4)) and 
                            np.any(self._data['face_colors'][:,3] < color_max))
        elif self._set['vertex']:
            transparency = (is_shape(self._data['vertex_colors'], (-1,4)) and
                            np.any(self._cache['vertex_colors'][:,3] < color_max))

        return self._cache.set(key   = 'transparency',
                               value = bool(transparency))

    def md5(self):
        result = self._data.md5()
        if self.mesh is not None:
            result += self.mesh.md5()
        return result

    @property
    def face_colors(self):
        stored = self._data['face_colors']
        if is_shape(stored, (len(self.mesh.faces), (3,4))):
            return stored
        log.debug('Returning default colors for faces.')
        self._data['face_colors'] = np.tile(DEFAULT_COLOR, 
                                            (len(self.mesh.faces), 1))
        return self._data['face_colors']

    @face_colors.setter
    def face_colors(self, values):
        values = np.asanyarray(values)
        if values.shape in ((3,), (4,)):
            # case where a single RGB/RGBa color has been passed to the setter
            # we apply this color to all faces 
            values = np.tile(values, (len(self.mesh.faces), 1))
        self._data['face_colors'] = rgba(values, dtype=self.dtype)

    @property
    def vertex_colors(self):
        cached = self._cache['vertex_colors']
        if is_shape(cached, (len(self.mesh.vertices), (3,4))):
            return cached

        log.debug('Vertex colors being generated from face colors')
        colors = face_to_vertex_color(self.mesh, self.face_colors)
        self._cache['vertex_colors'] = colors
        return colors
   
    @vertex_colors.setter
    def vertex_colors(self, values):
        self._cache['vertex_colors'] = rgba(values, dtype=self.dtype)

    def update_faces(self, mask):
        stored = self._data['face_colors']
        if not is_shape(stored, (-1,(3,4))):
            return 
        try: 
            self._data['face_colors'] = stored[mask]
        except: 
            log.warning('Face colors not updated', exc_info=True) 
            
    def update_vertices(self, mask):
        stored = self._data['vertex_colors']
        if not is_shape(stored, (-1, (3,4))):
            return
        try:    
            self._data['vertex_colors'] = stored[mask]
        except: 
            log.debug('Vertex colors not updated', exc_info=True)

    def subsets(self, faces_sequence):
        result = [VisualAttributes() for i in range(len(faces_sequence))]
        if self._set['face']:
            face = self._data['face_colors']
            for i, f in enumerate(faces_sequence):
                result[i].face_colors = face[list(f)]
        return np.array(result)

    def union(self, others):
        return visuals_union(np.append(self, others))

def _kwargs_to_color(mesh, **kwargs):
    '''
    Given a set of keyword arguments, see if any reference color
    in their name, and match the dimensions of the mesh.
    '''

    def pick_option(vf):
        if any(i is None for i in vf):
            return vf
        result = [None, None]
        signal = [i.ptp(axis=0).sum() for i in vf]
        signal_max = np.argmax(signal)
        result[signal_max] = vf[signal_max]
        return result
    def pick_color(sequence):
        if len(sequence) == 0:
            return None
        elif len(sequence) == 1:
            return sequence[0]
        else:
            signal = [i.ptp(axis=0).sum() for i in sequence]
            signal_max = np.argmax(signal)
            return sequence[signal_max]

    if mesh is None:
        result = [None, None]
        if 'face_colors' in kwargs:
            result[1] = np.asanyarray(kwargs['face_colors'])
        if 'vertex_colors' in kwargs:
            result[0] = np.asanyarray(kwargs['vertex_colors'])
        return result
        
    vertex = deque()
    face   = deque()
    
    for key in kwargs.keys():
        if not ('color' in key): 
            continue
        value = np.asanyarray(kwargs[key])
        if len(value) == len(mesh.vertices):
            vertex.append(value)
        elif len(value) == len(mesh.faces):
            face.append(value)
    return pick_option([pick_color(i) for i in [vertex, face]])

def visuals_union(visuals, *args):
    '''
    Concatenate two or more VisualAttribute objects.
    
    Arguments
    ----------
    visuals: VisualAttributes object, or array of same
    *args:   VisualAttributes objects
    
    Returns
    -----------
    concatenated: VisualAttributes object containing data from input visuals
    '''
    visuals = np.append(visuals, args)

    vertex_ok = True
    face_ok   = True
    face   = [None] * len(visuals)
    vertex = [None] * len(visuals)
    
    for i, v in enumerate(visuals):
        if face_ok:
            if v.mesh is None:
                # if the mesh is None, don't force a 
                # dimension check for the colors
                if util.is_shape(v._data['face_colors'], (-1,(3,4))):
                    face[i] = rgba(v._data['face_colors'])
                else:
                    face_ok = False
            else: 
                face[i] = rgba(v.face_colors)
        if vertex_ok:
            if v.mesh is None:
                if util.is_shape(v._data['vertex_colors'], (-1,(3,4))):
                    vertex[i] = rgba(v._data['vertex_colors'])
                else: 
                    vertex_ok = False
            else:
                vertex[i] = rgba(v.vertex_colors)
                
    color = {'face_colors'   : None,
             'vertex_colors' : None}
    if face_ok:   color['face_colors']   = np.vstack(face)
    if vertex_ok: color['vertex_colors'] = np.vstack(vertex)

    concatenated = VisualAttributes(**color)
    return concatenated
    
def color_to_float(color, dtype=None):
    color = np.asanyarray(color)
    if dtype is None:
        dtype = color.dtype
    else:
        color = color.astype(dtype)
    if dtype.kind in 'ui':
        signed = int(dtype.kind == 'i')
        color_max = float((2**((dtype.itemsize*8) - signed)) - 1)
        color = color.astype(np.float) / color_max
    return color

def rgba(colors, dtype=None):
    '''
    Convert an RGB color to an RGBA color.

    Arguments
    ----------
    colors: (n,[3|4]) set of RGB or RGBA colors
    
    Returns
    ----------
    colors: (n,4) set of RGBA colors
    '''
    if not is_sequence(colors): 
        return
    if dtype is None:
        dtype = COLOR_DTYPE
    colors = np.asanyarray(colors, dtype=dtype)
    if is_shape(colors, (-1,3)):
        opaque = (2**(np.dtype(dtype).itemsize * 8)) - 1
        colors = np.column_stack((colors,
                                  opaque * np.ones(len(colors)))).astype(dtype)
    return colors

def random_color(dtype=COLOR_DTYPE):
    '''
    Return a random RGB color using datatype specified.
    '''
    hue  = np.random.random() + .61803
    hue %= 1.0
    color = np.array(hsv_to_rgb(hue, .99, .99))
    if np.dtype(dtype).kind in 'iu':
        max_value = (2**(np.dtype(dtype).itemsize * 8)) - 1
        color    *= max_value
    color = np.append(color, max_value).astype(dtype)
    return color

def vertex_to_face_colors(vertex_colors, faces):
    face_colors = vertex_colors[faces].mean(axis=2).astype(vertex_colors.dtype)
    return face_colors

def face_to_vertex_color(mesh, face_colors, dtype=COLOR_DTYPE):
    '''
    Convert a set of face colors into a set of vertex colors.
    '''
    
    color_dim = np.shape(face_colors)[1]

    vertex_colors = np.zeros((len(mesh.vertices),3,color_dim))
    population    = np.zeros((len(mesh.vertices),3), dtype=np.bool)

    vertex_colors[[mesh.faces[:,0],0]] = face_colors
    vertex_colors[[mesh.faces[:,1],1]] = face_colors
    vertex_colors[[mesh.faces[:,2],2]] = face_colors

    population[[mesh.faces[:,0], 0]] = True
    population[[mesh.faces[:,1], 1]] = True
    population[[mesh.faces[:,2], 2]] = True

    # clip the population sum to 1, to avoid a division error in edge cases
    populated     = np.clip(population.sum(axis=1), 1, 3)
    vertex_colors = vertex_colors.sum(axis=1) / populated.reshape((-1,1))
    return vertex_colors.astype(dtype)



from .generic    import MeshScript
from ..resources import get_resource

from distutils.spawn import find_executable

_blender_executable = find_executable('blender')
_blender_template = get_resource('blender.py.template')

exists = _blender_executable is not None

def boolean(meshes, operation='difference'):
    if not exists:
        raise ValueError('No blender available!')
    operation = str.upper(operation)
    if operation == 'INTERSECTION': 
        operation = 'INTERSECT'

    script = _blender_template.replace('$operation', operation)
  
    with MeshScript(meshes = meshes, 
                    script = script) as blend:
        result = blend.run(_blender_executable + ' --background --python $script')

    # blender returns actively incorrect face normals
    result['face_normals'] = None
    return result





import numpy as np

from ..util      import is_binary_file

# define a numpy datatype for the STL file
_stl_dtype = np.dtype([('normals',    np.float32, (3)), 
                       ('vertices',   np.float32, (3,3)), 
                       ('attributes', np.uint16)])
_stl_dtype_header = np.dtype([('header', np.void, 80),
                             ('face_count', np.int32)])

def load_stl(file_obj, file_type=None):
    if is_binary_file(file_obj): return load_stl_binary(file_obj)
    else:                        return load_stl_ascii(file_obj)
        
def load_stl_binary(file_obj):
    '''
    Load a binary STL file into a trimesh object. 
    Uses a single main struct.unpack call, and is significantly faster
    than looping methods or ASCII STL. 
    '''

    header = np.fromstring(file_obj.read(84), dtype=_stl_dtype_header)

    # now we check the length from the header versus the length of the file
    # data_start should always be position 84, but hard coding that felt ugly
    data_start = file_obj.tell()
    # this seeks to the end of the file (position 0, relative to the end of the file 'whence=2')
    file_obj.seek(0, 2)
    # we save the location of the end of the file and seek back to where we started from
    data_end = file_obj.tell()
    file_obj.seek(data_start)
    
    # the binary format has a rigidly defined structure, and if the length
    # of the file doesn't match the header, the loaded version is almost
    # certainly going to be garbage. 
    data_ok = (data_end - data_start) == (header['face_count'] * _stl_dtype.itemsize)
   
    # this check is to see if this really is a binary STL file. 
    # if we don't do this and try to load a file that isn't structured properly 
    # the struct.unpack call uses 100% memory until the whole thing crashes, 
    # so it's much better to raise an exception here. 
    if not data_ok:
        raise ValueError('Binary STL has incorrect length in header!')
    
    # all of our vertices will be loaded in order due to the STL format, 
    # so faces are just sequential indices reshaped. 
    faces = np.arange(header['face_count'] * 3).reshape((-1,3))
    blob  = np.fromstring(file_obj.read(), dtype=_stl_dtype)
    
    result =  {'vertices'     : blob['vertices'].reshape((-1,3)),
               'face_normals' : blob['normals'].reshape((-1,3)),
               'faces'        : faces}
    return result
    
def load_stl_ascii(file_obj):
    '''
    Load an ASCII STL file.
    '''
    header = file_obj.readline()
    
    text = file_obj.read()
    if hasattr(text, 'decode'):
        text = text.decode('utf-8')
    text = text.lower().split('endsolid')[0]
    blob = np.array(text.split())

    # there are 21 'words' in each face
    face_len   = 21
    face_count = len(blob) / face_len
    if (len(blob) % face_len) != 0:
        raise ValueError('Incorrect number of values in STL file!')

    face_count   = int(face_count)
    # this offset is to be added to a fixed set of indices that is tiled
    offset       = face_len * np.arange(face_count).reshape((-1,1))
    normal_index = np.tile([2,3,4], (face_count, 1)) + offset
    vertex_index = np.tile([8,9,10,12,13,14,16,17,18], (face_count, 1)) + offset
    
    # faces are groups of three sequential vertices, as vertices are not references
    faces        = np.arange(face_count*3).reshape((-1,3))
    face_normals = blob[normal_index].astype(float)
    vertices     = blob[vertex_index.reshape((-1,3))].astype(float)

    return {'vertices'     : vertices,
            'faces'        : faces, 
            'face_normals' : face_normals}

def export_stl(mesh):
    '''
    Convert a Trimesh object into a binary STL file.

    Arguments
    ---------
    mesh: Trimesh object

    Returns
    ---------
    export: bytes, representing mesh in binary STL form
    '''
    header = np.zeros(1, dtype = _stl_dtype_header)
    header['face_count'] = len(mesh.faces)

    packed = np.zeros(len(mesh.faces), dtype=_stl_dtype)
    packed['normals']  = mesh.face_normals
    packed['vertices'] = mesh.triangles

    export  = header.tostring()
    export += packed.tostring()
    
    return export

_stl_loaders = {'stl':load_stl}






'''
trimesh.py

Library for importing and doing simple operations on triangular meshes.
'''

import numpy as np

from .constants import tol, log
from . import util

try:
    from scipy import spatial
except ImportError:
    log.warning('Scipy import failed!')

def convex_hull(obj, clean=True):
    '''
    Get a new Trimesh object representing the convex hull of the 
    current mesh. Requires scipy >.12.
    
    Argments
    --------
    clean: boolean, if True will fix normals and winding
           to be coherent (as qhull/scipy outputs are not)

    Returns
    --------
    convex: Trimesh object of convex hull of current mesh
    '''
    from .base import Trimesh
    if isinstance(obj, Trimesh):
        points = obj.vertices.view(np.ndarray)
    else:
        points = np.asanyarray(obj, dtype=np.float64)
        if not util.is_shape(points, (-1,3)):
            raise ValueError('Object must be Trimesh or (n,3) points!')
    c = spatial.ConvexHull(points.reshape((-1,3)),
                           qhull_options='QbB Pp')
    
    vid = np.sort(c.vertices)
    mask = np.zeros(len(c.points), dtype=np.int64)
    mask[vid] = np.arange(len(vid))
        
    faces    = mask[c.simplices]
    vertices = c.points[vid].copy()

    convex = Trimesh(vertices = vertices, 
                      faces    = faces,
                      process  = True)
    if clean:
        # the normals and triangle winding returned by scipy/qhull's
        # ConvexHull are apparently random, so we need to completely fix them
        convex.fix_normals()
    return convex

def is_convex(mesh, chunks=None):
    '''
    Test if a mesh is convex by projecting the vertices of 
    a triangle onto the normal of its adjacent face.
    
    Arguments
    ----------
    mesh: Trimesh object
    
    Returns
    ----------
    convex: bool, is the mesh convex or not
    '''
    chunk_block = 5e4
    if chunks is None:
        chunks = int(np.clip(len(mesh.faces) / chunk_block,1,10))

    # triangles from the second column of face adjacency
    triangles = mesh.triangles.copy()[mesh.face_adjacency[:,1]]
    # normals and origins from the first column of face adjacency
    normals = mesh.face_normals[mesh.face_adjacency[:,0]]
    origins = mesh.vertices[mesh.face_adjacency_edges[:,0]]

    # reshape and tile everything to be the same dimension
    triangles = triangles.reshape((-1,3))
    normals = np.tile(normals, (1,3)).reshape((-1,3))
    origins = np.tile(origins, (1,3)).reshape((-1,3))
    triangles -= origins

    # in non- convex meshes, we don't necessarily have to compute all 
    # dots of every face since we are looking for logical ALL
    for chunk_tri, chunk_norm in zip(np.array_split(triangles, chunks),
                                     np.array_split(normals,   chunks)):
        # project vertices of adjacent triangle onto normal
        # note that two of these are always going to be zero so we 
        # are doing more dot products than we really have to but finding
        # the index of the the third vertex through graph op is way slower 
        # than doing extra dot products. 
        # there is probably a clever way to use the winding to get this forfree
        dots = util.diagonal_dot(chunk_tri, chunk_norm)
        # if all projections are negative, or 'behind' the triangle
        # the mesh is convex
        if not bool((dots < tol.merge).all()):
            return False
    return True
        
    
def planar_hull(points, normal, origin=None, input_convex=False):
    '''
    Find the convex outline of a set of points projected to a plane.

    Arguments
    -----------
    points: (n,3) float, input points
    normal: (3) float vector, normal vector of plane
    origin: (3) float, location of plane origin
    input_convex: bool, if True we assume the input points are already from
                  a convex hull which provides a speedup. 

    Returns
    -----------
    hull_lines: (n,2,2) set of unordered line segments
    T:          (4,4) float, transformation matrix 
    '''
    from .points import project_to_plane

    if origin is None:
        origin = np.zeros(3)
    if not input_convex:
        pass
    planar, T = project_to_plane(points,
                                 plane_normal     = normal,
                                 plane_origin     = origin,
                                 return_planar    = False,
                                 return_transform = True)
    hull_edges = spatial.ConvexHull(planar[:,0:2]).simplices
    hull_lines = planar[hull_edges]
    planar_z = planar[:,2]
    height = np.array([planar_z.min(),
                       planar_z.max()])
    return hull_lines, T, height

def hull_points(obj):
    '''
    Try to extract a convex set of points from multiple input formats.

    Arguments
    ---------
    obj: Trimesh object
         (n,d) points
         (m,) Trimesh objects

    Returns
    --------
    points: (o,d) convex set of points
    '''
    if hasattr(obj, 'convex_hull_raw'):
        points = obj.convex_hull_raw.vertices
    elif util.is_sequence(obj):
        initial = np.asanyarray(obj)
        if len(initial.shape) != 2:
            raise ValueError('Points must be (n, dimension)!')
        hull   = spatial.ConvexHull(initial, qhull_options='QbB Pp')
        points = hull.points[hull.vertices]
    return points



import numpy as np

from ..constants import log

def load_assimp(file_obj, file_type=None):
    '''
    Use the assimp library to load a mesh, from a file object and type,
    or filename (if file_obj is a string)

    Assimp supports a huge number of mesh formats.

    Performance notes: in tests on binary STL pyassimp was ~10x 
    slower than the native loader included in this package. 
    This is probably due to their recursive prettifying of the data structure.
    
    Also, you need a very recent version of PyAssimp for this function to work 
    (the commit was merged into the assimp github master on roughly 9/5/2014)
    '''

    def LPMesh_to_Trimesh(lp):
        colors = (np.reshape(lp.colors, (-1,4))[:,0:3] * 255).astype(np.int)
        return {'vertices'       : lp.vertices,
                'vertex_normals' : lp.normals,
                'faces'          : lp.faces,
                'vertex_colors'  : colors}

    if not hasattr(file_obj, 'read'):
        # if there is no read attribute, we assume we've been passed a file name
        file_type = (str(file_obj).split('.')[-1]).lower()
        file_obj  = open(file_obj, 'rb')

    scene  = pyassimp.load(file_obj, file_type=file_type)
    meshes = list(map(LPMesh_to_Trimesh, scene.meshes))
    pyassimp.release(scene)

    if len(meshes) == 1: 
        return meshes[0]
    return meshes
 
_assimp_loaders = {}
try: 
    import pyassimp
    if hasattr(pyassimp, 'available_formats'):
        _assimp_formats = [i.lower() for i in pyassimp.available_formats()]
    else: 
        log.warning('Older version of assimp detected, using hardcoded format list')
        _assimp_formats = ['dae', 'blend', '3ds', 'ase',  'obj', 
                           'ifc', 'xgl',   'zgl', 'ply',  'lwo',
                           'lxo', 'x',     'ac',  'ms3d', 'cob', 'scn']
    _assimp_loaders.update(zip(_assimp_formats,
                               [load_assimp]*len(_assimp_formats)))
except:
    log.warning('pyassimp unavailable, using only native loaders')



import numpy as np

from ..util         import three_dimensionalize, euclidean, unitize
from ..constants    import log
from ..constants    import tol_path as tol
from ..constants    import res_path as res
from .intersections import line_line



def arc_center(points):
    '''
    Given three points of an arc, find the center, radius, normal, and angle.

    This uses the fact that the intersection of the perpendicular
    bisectors of the segments between the control points is the center of the arc.

    Arguments
    ---------
    points: (3,d) list of points where (d in [2,3])
    
    Returns
    ---------
    center:       (d) point of the center of the arc
    radius:       float, radius of the arc
    plane_normal: (3) vector of the plane normal. 
    angle:        float, angle swept by the arc
    '''
    #it's a lot easier to treat 2D as 3D with a zero Z value
    is_2D, points = three_dimensionalize(points, return_2D = True)

    #find the two edge vectors of the triangle
    edge_direction = np.diff(points, axis=0)
    edge_midpoints = (edge_direction*.5) + points[0:2]

    #three points define a plane, so we find its normal vector
    plane_normal         = unitize(np.cross(*edge_direction[::-1]))
    vector_edge          = unitize(edge_direction)
    vector_perpendicular = unitize(np.cross(vector_edge, plane_normal))

    intersects, center   = line_line(edge_midpoints, vector_perpendicular)

    if not intersects:
        raise ValueError('Segments do not intersect!')

    radius  = euclidean(points[0], center)
    vector  = unitize(points - center)
    angle   = np.arccos(np.clip(np.dot(*vector[[0,2]]), -1.0, 1.0))
    large_arc = (abs(angle) > tol.zero and 
                 np.dot(*edge_direction) < 0.0)
    if large_arc:
        angle = (np.pi*2) - angle

    angles = np.arctan2(*vector[:,0:2].T[::-1]) + np.pi*2
    angles_sorted = np.sort(angles[[0,2]])
    reverse  = angles_sorted[0] < angles[1] < angles_sorted[1]
    angles_sorted = angles_sorted[::(1 - int(not reverse)*2)]

    result = {'center' : center[:(3-is_2D)], 
              'radius' : radius, 
              'normal' : plane_normal, 
              'span'   : angle, 
              'angles' : angles_sorted}
    return result

def discretize_arc(points, close = False, scale=1.0):
    '''
    Returns a version of a three point arc consisting of line segments

    Arguments
    ---------
    points: (n, d) points on the arc where d in [2,3]
    close:  boolean, if True close the arc (circle)

    Returns
    ---------
    discrete: (m, d)  
    points: either (3,3) or (3,2) of points for arc going from 
            points[0] to points[2], going through control point points[1]
    '''
    two_dimensional, points = three_dimensionalize(points, return_2D = True)
    center_info = arc_center(points)
    center, R, N, angle = (center_info['center'],
                           center_info['radius'],
                           center_info['normal'],
                           center_info['span'])
    if close: angle = np.pi * 2
    
    #the number of facets, based on the angle critera
    count_a = angle / res.seg_angle
    count_l = ((R*angle)) / (res.seg_frac * scale)
    

    count = np.max([count_a, count_l])
    # force at LEAST 4 points for the arc, otherwise the endpoints will diverge
    count = np.clip(count, 4, np.inf)
    count = int(np.ceil(count))

    V1 = unitize(points[0] - center)
    V2 = unitize(np.cross(-N, V1))
    t  = np.linspace(0, angle, count)

    discrete  = np.tile(center, (count, 1))
    discrete += R * np.cos(t).reshape((-1,1))*np.tile(V1, (count, 1)) 
    discrete += R * np.sin(t).reshape((-1,1))*np.tile(V2, (count, 1))

    if not close:
        arc_dist = np.linalg.norm(points[[0,-1]]-discrete[[0,-1]], axis=1) 
        arc_ok   = (arc_dist < tol.merge).all()
        if not arc_ok:
            log.warn('Failed to discretize arc (endpoint distance %s)', str(arc_dist))
            log.warn('Failed arc points: %s', str(points))
            raise ValueError('Arc endpoints diverging!')
    discrete  = discrete[:,0:(3-two_dimensional)]

    return discrete

def arc_tangents(points):
    '''
    returns tangent vectors for points
    '''
    two_dimensional, points = three_dimensionalize(points, return_2D=True)
    center, R, N, angle     = arc_center(points)
    vectors  = points - center
    tangents = unitize(np.cross(vectors, N))
    return tangents[:,0:(3-two_dimensional)]

def arc_offset(points, distance):
    two_dimensional, points = three_dimensionalize(points)
    center, R, N, angle     = arc_center(points)
    vectors    = unitize(points - center)
    new_points = center + vectors*distance
    return new_points[:,0:(3-two_dimensional)]

def angles_to_threepoint(angles, center, radius):
    if angles[1] < angles[0]: angles[1] += np.pi*2
    angles = [angles[0], np.mean(angles), angles[1]]
    planar = np.column_stack((np.cos(angles), np.sin(angles)))*radius
    return planar + center



import numpy as np

from ...constants import log
from ..entities  import Line, Arc, Bezier

from collections import deque
from xml.dom.minidom import parseString as parse_xml

try:     
    from svg.path import parse_path
except:
    log.warning('SVG path loading unavailable!')

def svg_to_path(file_obj, file_type=None):
    def complex_to_float(values):
        return np.array([[i.real, i.imag] for i in values])

    def load_line(svg_line):
        points = complex_to_float([svg_line.point(0.0),
                                   svg_line.point(1.0)])
        if not starting: points[0] = vertices[-1]
        entities.append(Line(np.arange(2)+len(vertices)))
        vertices.extend(points)

    def load_arc(svg_arc):
        points = complex_to_float([svg_arc.start, 
                                   svg_arc.point(.5), 
                                   svg_arc.end])
        if not starting: points[0] = vertices[-1]
        entities.append(Arc(np.arange(3)+len(vertices)))
        vertices.extend(points)
    def load_quadratic(svg_quadratic):
        points = complex_to_float([svg_quadratic.start, 
                                   svg_quadratic.control, 
                                   svg_quadratic.end])
        if not starting: points[0] = vertices[-1]
        entities.append(Bezier(np.arange(3)+len(vertices)))
        vertices.extend(points)
    def load_cubic(svg_cubic):
        points = complex_to_float([svg_cubic.start, 
                                   svg_cubic.control1, 
                                   svg_cubic.control2, 
                                   svg_cubic.end])
        if not starting: points[0] = vertices[-1]
        entities.append(Bezier(np.arange(4)+len(vertices)))
        vertices.extend(points)

    # first, we grab all of the path strings from the xml file
    xml   = parse_xml(file_obj.read())
    paths = [p.attributes['d'].value for p in xml.getElementsByTagName('path')]

    entities = deque()
    vertices = deque()  
    loaders  = {'Arc'             : load_arc,
                'Line'            : load_line,
                'CubicBezier'     : load_cubic,
                'QuadraticBezier' : load_quadratic}

    for svg_string in paths:
        starting = True
        for svg_entity in parse_path(svg_string):
            loaders[svg_entity.__class__.__name__](svg_entity)

    return {'entities' : np.array(entities),
            'vertices' : np.array(vertices)}



import numpy as np

from .grouping import unique_rows

def subdivide(mesh, face_index=None):
    '''
    Subdivide a mesh into smaller triangles.

    Arguments
    ----------
    mesh: Trimesh object
    face_index: faces to subdivide.
                if None: all faces of mesh will be subdivided
                if (n,) int array of indices: only specified faces will be 
                   subdivided. Note that in this case the mesh will generally
                   no longer be manifold, as the additional vertex on the midpoint
                   will not be used by the adjacent faces to the faces specified,
                   and an additional postprocessing step will be required to 
                   make resulting mesh watertight


    '''
    if face_index is None:
        face_index = np.arange(len(mesh.faces))
    else:
        face_index = np.asanyarray(face_index, dtype=np.int64)

    # the (c,3) int set of vertex indices
    faces = mesh.faces[face_index]
    # the (c, 3, 3) float set of points in the triangles
    triangles = mesh.triangles[face_index]
    # the 3 midpoints of each triangle edge vstacked to a (3*c, 3) float
    mid = np.vstack([triangles[:,g,:].mean(axis=1) for g in [[0,1], 
                                                             [1,2], 
                                                             [2,0]]])
    mid_idx = (np.arange(len(face_index)*3)).reshape((3,-1)).T
    # for adjacent faces we are going to be generating the same midpoint
    # twice, so we handle it here by finding the unique vertices
    unique,inverse = unique_rows(mid)
    
    mid = mid[unique]
    mid_idx = inverse[mid_idx] + len(mesh.vertices)
    # the new faces, with correct winding
    f = np.column_stack([faces[:,0], mid_idx[:,0], mid_idx[:,2],
                         mid_idx[:,0], faces[:,1], mid_idx[:,1],
                         mid_idx[:,2], mid_idx[:,1], faces[:,2],
                         mid_idx[:,0], mid_idx[:,1], mid_idx[:,2],]).reshape((-1,3))
    # add the 3 new faces per old face
    new_faces = np.vstack((mesh.faces, f[len(face_index):]))
    # replace the old face with a smaller face
    new_faces[face_index] = f[:len(face_index)]

    mesh.vertices = np.vstack((mesh.vertices, mid))
    mesh.faces = new_faces
 
 



import numpy as np

def sample_surface(mesh, count):
    '''
    Sample the surface of a mesh, returning the specified number of points

    For individual triangle sampling uses this method:
    http://mathworld.wolfram.com/TrianglePointPicking.html

    Arguments
    ---------
    mesh: Trimesh object
    count: number of points to return
    
    Returns
    ---------
    samples: (count,3) points in space on the surface of mesh
         
    '''

    #len(mesh.faces) float array of the areas of each face of the mesh
    area     = mesh.area_faces
    # total area (float)
    area_sum = np.sum(area)
    # cumulative area (len(mesh.faces)) 
    area_cum = np.cumsum(area)
    face_pick  = np.random.random(count) * area_sum
    face_index = np.searchsorted(area_cum, face_pick)

    # pull triangles into the form of an origin + 2 vectors
    tri_origins  = mesh.triangles[:,0]
    tri_vectors  = mesh.triangles[:,1:].copy()
    tri_vectors -= np.tile(tri_origins, (1,2)).reshape((-1,2,3))

    # pull the vectors for the faces we are going to sample from
    tri_origins = tri_origins[face_index]
    tri_vectors = tri_vectors[face_index]

    # randomly generate two 0-1 scalar components to multiply edge vectors by
    random_lengths = np.random.random((len(tri_vectors), 2, 1))

    # points will be distributed on a quadrilateral if we use 2 0-1 samples
    # if the two scalar components sum less than 1.0 the point will be
    # inside the triangle, so we find vectors longer than 1.0 and
    # transform them to be inside the triangle
    random_test = random_lengths.sum(axis=1).reshape(-1) > 1.0
    random_lengths[random_test] -= 1.0
    random_lengths = np.abs(random_lengths)

    # multiply triangle edge vectors by the random lengths and sum
    sample_vector = (tri_vectors*random_lengths).sum(axis=1)

    # finally, offset by the origin to generate
    # (n,3) points in space on the triangle
    samples = sample_vector + tri_origins
    
    return samples

def sample_volume(mesh, count):
    '''
    
    '''
    points = (np.random.random((count, 3))*mesh.extents) + mesh.bounds[0]
    contained = mesh.contains(points)
    samples = points[contained]
    return samples

def sample_surface_even(mesh, count):
    '''
    Sample the surface of a mesh, returning samples which are 
    approximately evenly spaced. 
    '''
    from .points import remove_close

    radius = np.sqrt(mesh.area / (2*count))
    samples = sample_surface(mesh, count*5)
    result = remove_close(samples, radius)
    return result



import pyglet
import pyglet.gl as gl
import numpy as np

import platform

from collections import deque

from ..transformations import Arcball

#smooth only when fewer faces than this
_SMOOTH_MAX_FACES = 100000


class SceneViewer(pyglet.window.Window):
    def __init__(self, 
                 scene, 
                 smooth = None,
                 save_image = None,
                 flags = None,
                 resolution = (640,480)):

        self.scene = scene
        self.scene._redraw = self._redraw
        
        self.reset_view(flags=flags)

        visible = (save_image is None) or (platform.system() != 'Linux')
        width, height = resolution

        try:
            conf = gl.Config(sample_buffers = 1,
                             samples        = 4,
                             depth_size     = 16,
                             double_buffer  = True)
            super(SceneViewer, self).__init__(config=conf, 
                                              visible=visible, 
                                              resizable=True,
                                              width=width,
                                              height=height)
        except pyglet.window.NoSuchConfigException:
            conf = gl.Config(double_buffer=True)
            super(SceneViewer, self).__init__(config = conf,
                                              resizable=True,
                                              visible = visible,
                                              width=width,
                                              height=height)

        self.batch = pyglet.graphics.Batch()
        self._img  = save_image

        self.vertex_list     = {}
        self.vertex_list_md5 = {}

        for name, mesh in scene.meshes.items():
            self._add_mesh(name_mesh = name, 
                           mesh      = mesh, 
                           smooth    = smooth)
        self.init_gl()
        self.set_size(*resolution)
        self.update_flags()
        pyglet.app.run()

    def _redraw(self):
        self.on_draw()

    def _update_meshes(self):
        for name, mesh in self.scene.meshes.items():
            md5 = mesh.md5() + mesh.visual.md5()
            if self.vertex_list_md5[name] != md5:
                self._add_mesh(name, mesh)

    def _add_mesh(self, name_mesh, mesh, smooth=None):    
        if smooth is None:
            smooth = len(mesh.faces) < _SMOOTH_MAX_FACES
        if smooth:
            display = mesh.smoothed()
        else:
            display = mesh.copy()
            display.unmerge_vertices()
        self.vertex_list[name_mesh] = self.batch.add_indexed(*mesh_to_vertex_list(display))
        self.vertex_list_md5[name_mesh] = mesh.md5() + mesh.visual.md5()

    def reset_view(self, flags=None):
        '''
        Set view to base.
        '''
        self.view = {'wireframe'   : False,
                     'cull'        : True,
                     'translation' : np.zeros(3),
                     'center'      : self.scene.centroid,
                     'scale'       : self.scene.scale,
                     'ball'        : Arcball()}
        if isinstance(flags, dict):
            for k,v in flags.items():
                if k in self.view:
                    self.view[k] = v
        self.update_flags()

    def init_gl(self):
        gl.glClearColor(.93, .93, 1, 1)
        #glColor3f(1, 0, 0)

        gl.glEnable(gl.GL_DEPTH_TEST)
        gl.glEnable(gl.GL_CULL_FACE)
        gl.glEnable(gl.GL_LIGHTING)
        gl.glEnable(gl.GL_LIGHT0)
        gl.glEnable(gl.GL_LIGHT1)

        gl.glLightfv(gl.GL_LIGHT0, gl.GL_POSITION, _gl_vector(.5, .5, 1, 0))
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_SPECULAR, _gl_vector(.5, .5, 1, 1))
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_DIFFUSE, _gl_vector(1, 1, 1, 1))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_POSITION, _gl_vector(1, 0, .5, 0))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_DIFFUSE, _gl_vector(.5, .5, .5, 1))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_SPECULAR, _gl_vector(1, 1, 1, 1))

        gl.glColorMaterial(gl.GL_FRONT_AND_BACK, gl.GL_AMBIENT_AND_DIFFUSE)
        gl.glEnable(gl.GL_COLOR_MATERIAL)
        gl.glShadeModel(gl.GL_SMOOTH)

        gl.glMaterialfv(gl.GL_FRONT, gl.GL_AMBIENT, _gl_vector(0.192250, 0.192250, 0.192250))
        gl.glMaterialfv(gl.GL_FRONT, gl.GL_DIFFUSE, _gl_vector(0.507540, 0.507540, 0.507540))
        gl.glMaterialfv(gl.GL_FRONT, gl.GL_SPECULAR, _gl_vector(.5082730,.5082730,.5082730))

        gl.glMaterialf(gl.GL_FRONT, gl.GL_SHININESS, .4 * 128.0);

        gl.glEnable(gl.GL_BLEND) 
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA) 


    def toggle_culling(self):
        self.view['cull'] = not self.view['cull']
        self.update_flags()

    def toggle_wireframe(self):
        self.view['wireframe'] = not self.view['wireframe']
        self.update_flags()

    def update_flags(self):
        if self.view['wireframe']: 
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
        else:
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

        if self.view['cull']:
            gl.glEnable(gl.GL_CULL_FACE)
        else:
            gl.glDisable(gl.GL_CULL_FACE)

    def on_resize(self, width, height):
        gl.glViewport(0, 0, width, height)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        gl.gluPerspective(60., width / float(height), .01, self.scene.scale*5.0)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        self.view['ball'].place([width/2, height/2], (width+height)/2)
        
    def on_mouse_press(self, x, y, buttons, modifiers):
        self.view['ball'].down([x,-y])
        
    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        delta = np.array([dx, dy], dtype=np.float) / [self.width, self.height]

        #left mouse button, with control key down (pan)
        if ((buttons == pyglet.window.mouse.LEFT) and 
            (modifiers & pyglet.window.key.MOD_CTRL)):
            self.view['translation'][0:2] += delta

        #left mouse button, no modifier keys pressed (rotate)
        elif (buttons == pyglet.window.mouse.LEFT):
            self.view['ball'].drag([x,-y])

    def on_mouse_scroll(self, x, y, dx, dy):
        self.view['translation'][2] += float(dy) / self.height
        
    def on_key_press(self, symbol, modifiers):
        if symbol == pyglet.window.key.W:
            self.toggle_wireframe()
        elif symbol == pyglet.window.key.Z:
            self.reset_view()
        elif symbol == pyglet.window.key.C:
            self.toggle_culling()

    def on_draw(self):
        self._update_meshes()
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glLoadIdentity()

        # pull the new camera transform from the scene
        transform_camera = self.scene.transforms['camera']
        # apply the camera transform to the matrix stack
        gl.glMultMatrixf(_gl_matrix(transform_camera))
        
        # dragging the mouse moves the view transform (but doesn't alter the scene)
        transform_view = _view_transform(self.view)
        gl.glMultMatrixf(_gl_matrix(transform_view))

        # we want to render fully opaque objects first,
        # followed by objects which have transparency
        items = deque(self.scene.nodes.items())
        count_original = len(items)
        count = -1
        
        while len(items) > 0:
            count += 1
            item = items.popleft()
            name_node, name_mesh = item

            # if the flag isn't defined, this will be None
            # by checking False explicitly, it makes the default
            # behaviour to render meshes with no flag defined. 
            if self.node_flag(name_node, 'visible') == False:
                continue
                
            if self.scene.meshes[name_mesh].visual.transparency:
                # put the current item onto the back of the queue
                if count < count_original:
                    items.append(item)
                    continue

            transform = self.scene.transforms[name_node]
            # add a new matrix to the model stack
            gl.glPushMatrix()
            # transform by the nodes transform
            gl.glMultMatrixf(_gl_matrix(transform))
            # draw the mesh with its transform applied
            self.vertex_list[name_mesh].draw(mode=gl.GL_TRIANGLES)
            # pop the matrix stack as we drew what we needed to draw
            gl.glPopMatrix()

    def node_flag(self, node, flag):
        if (hasattr(self.scene, 'flags') and
            node in self.scene.flags and
            flag in self.scene.flags[node]):
            return self.scene.flags[node][flag]
        return None
        
    def save_image(self, filename):
        colorbuffer = pyglet.image.get_buffer_manager().get_color_buffer()
        colorbuffer.save(filename)

    def flip(self):
        '''
        This function is the last thing executed in the event loop,
        so if we want to close the window (self) here is the place to do it.
        '''
        super(self.__class__, self).flip()

        if self._img is not None:
            self.save_image(self._img)
            self.close()

def _view_transform(view):
    '''
    Given a dictionary containing view parameters,
    calculate a transformation matrix. 
    '''
    transform         = view['ball'].matrix()
    transform[0:3,3]  = view['center']
    transform[0:3,3] -= np.dot(transform[0:3,0:3], view['center'])
    transform[0:3,3] += view['translation'] * view['scale'] * 5.0
    return transform

def mesh_to_vertex_list(mesh, group=None):
    '''
    Convert a Trimesh object to arguments for an 
    indexed vertex list constructor. 
    '''
    mesh.visual.choose()
    
    normals  = mesh.vertex_normals.reshape(-1).tolist()
    colors   = mesh.visual.vertex_colors.reshape(-1).tolist()
    faces    = mesh.faces.reshape(-1).tolist()
    vertices = mesh.vertices.reshape(-1).tolist()
    
    color_dimension = mesh.visual.vertex_colors.shape[1]
    color_type = 'c' + str(color_dimension) + 'B/static'

    args = (len(mesh.vertices), # number of vertices
            gl.GL_TRIANGLES,       # mode 
            group,              # group
            faces,              # indices 
            ('v3f/static', vertices),
            ('n3f/static', normals),
            (color_type,   colors))
    return args
    
def _gl_matrix(array):
    '''
    Convert a sane numpy transformation matrix (row major, (4,4))
    to an stupid GLfloat transformation matrix (column major, (16,))
    '''
    a = np.array(array).T.reshape(-1)
    return (gl.GLfloat * len(a))(*a)

def _gl_vector(array, *args):
    '''
    Convert an array and an optional set of args into a flat vector of GLfloat
    '''
    array = np.array(array)
    if len(args) > 0:
        array = np.append(array, args)
    vector = (gl.GLfloat * len(array))(*array)
    return vector



import numpy as np

from ..grouping        import group_rows
from ..transformations import rotation_matrix, transform_points
from .transforms       import TransformForest

from .. import util


from collections import deque

class Scene:
    '''
    A simple scene graph which can be rendered directly via pyglet/openGL,
    or through other endpoints such as a raytracer. 

    Meshes and lights are added by name, which can then be moved by updating
    transform in the transform tree. 
    '''

    def __init__(self, 
                 node       = None,
                 base_frame ='world'):

        # instance name : mesh name
        self.nodes = {}
        self._cache = util.Cache(id_function=self.md5)

        # mesh name : Trimesh object
        self.meshes     = {}
        self.flags      = {}
        self.transforms = TransformForest(base_frame = base_frame)

        if node is not None:
            self.add_mesh(node)
            self.set_camera()
            
        self._cache.id_set()

    def add_mesh(self, mesh):
        '''
        Add a mesh to the scene.

        If the mesh has multiple transforms defined in its metadata, 
        a new instance of the mesh will be created at each transform. 
        '''        
        if util.is_sequence(mesh):
            for i in mesh:
                self.add_mesh(i)
            return

        if 'name' in mesh.metadata: 
            name_mesh = mesh.metadata['name']
        else:
            name_mesh = 'mesh_' + str(len(self.meshes))

        self.meshes[name_mesh] = mesh

        if 'transforms' in mesh.metadata:
            transforms = np.array(mesh.metadata['transforms'])
        else:
            transforms = np.eye(4).reshape((-1,4,4))

        for i, transform in enumerate(transforms):
            name_node = name_mesh + '_' + str(i)
            self.nodes[name_node] = name_mesh
            self.flags[name_node] = {'visible':True}
            self.transforms.update(frame_to = name_node, 
                                   matrix   = transform)

    def md5(self):
        '''
        MD5 of scene, which will change when meshes or transforms are changed
        '''
        mesh_hash = util.md5_object(np.sort([hash(i) for i in self.meshes.values()]))
        result = mesh_hash + self.transforms.md5()
        return result

    @util.cache_decorator
    def bounds(self):
        '''
        Return the overall bounding box of the scene.

        Returns
        --------
        bounds: (2,3) float points for min, max corner
        '''
        corners = deque()
        for instance, mesh_name in self.nodes.items():
            transform = self.transforms.get(instance)
            corners.append(transform_points(self.meshes[mesh_name].bounds, 
                                            transform))
        corners = np.vstack(corners)
        bounds  = np.array([corners.min(axis=0), 
                            corners.max(axis=0)])
        return bounds

    @util.cache_decorator
    def extents(self):
        return np.diff(self.bounds, axis=0).reshape(-1)

    @util.cache_decorator
    def scale(self):
        '''
        The approximate scale of the mesh
        
        Returns
        -----------
        scale: float, the mean of the bounding box edge lengths
        '''
        return self.extents.mean()

    @util.cache_decorator
    def centroid(self):
        '''
        Return the center of the bounding box for the scene.

        Returns
        --------
        centroid: (3) float point for center of bounding box
        '''
        centroid = np.mean(self.bounds, axis=0)
        return centroid

    def duplicate_nodes(self):
        '''
        Return a sequence of node keys of identical meshes.

        Will combine meshes duplicated by copying in space with different keys in 
        self.meshes, as well as meshes repeated by self.nodes.

        Returns
        -----------
        duplicates: (m) sequence of keys to self.nodes that represent identical geometry 
        '''
        mesh_ids  = {k : m.identifier for k, m in self.meshes.items()}
        
        node_keys = np.array(list(self.nodes.keys()))
        node_ids  = [mesh_ids[v] for v in self.nodes.values()]
        
        node_groups = group_rows(node_ids, digits=1)

        duplicates  = [np.sort(node_keys[g]).tolist() for g in node_groups]
        return duplicates

    def set_camera(self, angles=None, distance=None, center=None):
        if center is None:
            center = self.centroid
        if distance is None:
            distance = self.extents.max()
        if angles is None:
            angles = np.zeros(3)

        translation        = np.eye(4)
        translation[0:3,3] = center
        translation[2][3] += distance*1.5

        transform = np.dot(rotation_matrix(angles[0], [1,0,0], point=center),
                           rotation_matrix(angles[1], [0,1,0], point=center))
        transform = np.dot(transform, translation)

        self.transforms.update(frame_from = 'camera', 
                               frame_to   = self.transforms.base_frame,
                               matrix     = transform)

    def dump(self):
        '''
        Append all meshes in scene to a list of meshes.
        '''
        result = deque()
        for node_id, mesh_id in self.nodes.items():
            transform = self.transforms.get(node_id)
            current   = self.meshes[mesh_id].copy()
            current.apply_transform(transform)
            result.append(current)
        return np.array(result)

    def export(self, file_type='dict64'):
        '''
        Export a snapshot of the current scene.

        Arguments
        ----------
        file_type: what encoding to use for meshes
                   ie: dict, dict64, stl
        
        Returns
        ----------
        export: dict with keys:
                meshes: list of meshes, encoded as per file_type
                transforms: edge list of transforms, eg:
                             ((u, v, {'matrix' : np.eye(4)}))
        '''
        export = {'transforms' : self.transforms.export(),
                  'nodes'      : self.nodes,
                  'meshes'     : {},
                  'scene_cache' : self._cache.cache}
        # if the mesh has an export method use it, otherwise put the mesh
        # itself into the export object
        for node, mesh in self.meshes.items():
            if hasattr(mesh, 'export'): 
                export['meshes'][node] = mesh.export(file_type=file_type)
            else: 
                export['meshes'][node] = mesh
        return export
        
    def save_image(self, file_obj, resolution=(1024,768), **kwargs):
        from .viewer import SceneViewer
        SceneViewer(self, 
                    save_image = file_obj, 
                    resolution = resolution, 
                    **kwargs)

    def explode(self, vector=[0.0,0.0,1.0], origin=None):
        '''
        Explode a scene around a point and vector.
        '''
        if origin is None:
            origin = self.centroid
        centroids = np.array([self.meshes[i].centroid for i in self.nodes.values()])

        if np.shape(vector) == (3,):
            vectors = np.tile(vector, (len(centroids), 1))
            projected = np.dot(vector, (centroids-origin).T)
            offsets = vectors * projected.reshape((-1,1))
        elif isinstance(vector, float):
            offsets = (centroids-origin) * vector
        else:
            raise ValueError('Explode vector must by (3,) or float')

        for offset, node_key in zip(offsets, self.nodes.keys()):
            current = self.transforms[node_key]
            current[0:3,3] += offset
            self.transforms[node_key] = current

    def show(self, block=True, **kwargs):
        # this imports pyglet, and will raise an ImportError
        # if pyglet is not available 
        from .viewer import SceneViewer

        def viewer():
            SceneViewer(self, **kwargs)

        if block:
            viewer()
        else:
            from threading import Thread
            Thread(target = viewer, kwargs=kwargs).start()



import numpy as np
import json

from .. import util

def load_off(file_obj, file_type=None):
    '''
    Load an OFF file into the kwargs for a Trimesh constructor


    Arguments
    ----------
    file_obj: file object containing an OFF file
    file_type: not used
    
    Returns
    ----------
    loaded: dict with kwargs for Trimesh constructor (vertices, faces)
 
    '''
    header_string = file_obj.readline()
    if hasattr(header_string, 'decode'):
        header_string = header_string.decode('utf-8')
    header_string = header_string.strip()

    if not header_string == 'OFF': 
        raise NameError('Not an OFF file! Header was ' + header_string)

    header  = np.array(file_obj.readline().split()).astype(int)
    blob    = np.array(file_obj.read().split())
    data_ok = np.sum(header * [3,4,0]) == len(blob)    
    if not data_ok: raise NameError('Incorrect number of vertices or faces!')

    vertices = blob[0:(header[0]*3)].astype(float).reshape((-1,3))
    faces    = blob[(header[0]*3):].astype(int).reshape((-1,4))[:,1:]
  
    return {'vertices' : vertices,
            'faces'    : faces}

def load_wavefront(file_obj, file_type=None):
    '''
    Loads an ascii Wavefront OBJ file_obj into kwargs for the Trimesh constructor.

    Discards texture normals and vertex color information.


    Arguments
    ----------
    file_obj: file object containing a wavefront file
    file_type: not used
    
    Returns
    ----------
    loaded: dict with kwargs for Trimesh constructor (vertices, faces)
    '''
    data     = np.array(file_obj.read().split())
    data_str = data.astype(str)
    
    # find the locations of keys, then find the proceeding values
    vid = np.nonzero(data_str == 'v')[0].reshape((-1,1))  + np.arange(3) + 1
    nid = np.nonzero(data_str == 'vn')[0].reshape((-1,1)) + np.arange(3) + 1
    fid = np.nonzero(data_str == 'f')[0].reshape((-1,1))  + np.arange(3) + 1
    
    # if we wanted to use the texture/vertex normals, we could slice this differently.
    faces = np.array([i.split(b'/') for i in data[fid].reshape(-1)])[:,0].reshape((-1,3))
    # wavefront has 1- indexed faces, as opposed to 0- indexed
    faces = faces.astype(int) - 1
    loaded = {'vertices'       : data[vid].astype(float),
              'vertex_normals' : data[nid].astype(float),
              'faces'          : faces}
    return loaded

def load_msgpack(blob, file_type=None):
    '''
    Load a dict packed with msgpack into kwargs for Trimesh constructor

    Arguments
    ----------
    blob: msgpack packed dict with keys for 'vertices' and 'faces'
    file_type: not used

    Returns
    ----------
    loaded: kwargs for Trimesh constructor (aka mesh=trimesh.Trimesh(**loaded))
    '''

    import msgpack
    if hasattr(blob, 'read'):
        data = msgpack.load(blob)
    else:
        data = msgpack.loads(blob)
    loaded = load_dict(data)
    return loaded

def load_dict(data, file_type=None):
    '''
    Load multiple input types into kwargs for a Trimesh constructor.
    Tries to extract keys ['faces', 'vertices', 'face_normals', 'vertex_normals'].

    Arguments
    ----------
    data: accepts multiple forms
          -dict: has keys for vertices and faces as (n,3) numpy arrays
          -dict: has keys for vertices/faces (n,3) arrays encoded as dicts/base64
                 with trimesh.util.array_to_encoded/trimesh.util.encoded_to_array
          -str:  json blob as dict with either straight array or base64 values
          -file object: json blob of dict
    file_type: not used

    Returns
    -----------
    loaded: dict with keys 
            -vertices: (n,3) float
            -faces:    (n,3) int
            -face_normals: (n,3) float (optional)
    '''
    if data is None:
        raise ValueError('data passed to load_dict was None!')
    if util.is_string(data):
        if not '{' in data:
            raise ValueError('Object is not a JSON encoded dictionary!')
        data = json.loads(data.decode('utf-8'))
    elif util.is_file(data):
        data = json.load(data)

    # what shape should the data be to be usable
    mesh_data = {'vertices'       : (-1,3),
                 'faces'          : (-1, (3,4)),
                 'face_normals'   : (-1,3),
                 'vertex_normals' : (-1,3)}
    # now go through data structure and if anything is encoded as base64
    # pull it back into numpy arrays
    if util.is_dict(data):
        loaded = {}
        data = util.decode_keys(data, 'utf-8')
        for key, shape in mesh_data.items():
            if key in data: 
                loaded[key] = util.encoded_to_array(data[key])
                if not util.is_shape(loaded[key], shape):
                    raise ValueError('Shape of %s is %s, not %s!',
                                     key,
                                     str(loaded[key].shape),
                                     str(shape))
        if len(key) == 0:
            raise ValueError('Unable to extract any mesh data!')
        return loaded
    else:
        raise ValueError('%s object passed to dict loader!',
                         data.__class__.__name__)

_misc_loaders = {'obj'     : load_wavefront,
                 'off'     : load_off,
                 'dict'    : load_dict,
                 'dict64'  : load_dict,
                 'json'    : load_dict,
                 'msgpack' : load_msgpack}



import numpy as np

from . import convex
from .constants import log

try:
    from scipy          import spatial
    from scipy.optimize import leastsq
except ImportError:
    log.warning('No scipy!')

def minimum_nsphere(obj):
    '''
    Compute the minimum n- sphere for a mesh or a set of points.

    Uses the fact that the minimum n- sphere will be centered at one of
    the vertices of the furthest site voronoi diagram, which is n*log(n)
    but should be pretty fast due to using the scipy/qhull implementations
    of convex hulls and voronoi diagrams.
   
    Arguments
    ----------
    obj: Trimesh object OR
         (n,d) float, set of points

    Returns
    ----------
    center: (d) float, center of n- sphere
    radius: float, radius of n-sphere
    '''
    # reduce the input points or mesh to the vertices of the convex hull
    # since we are computing the furthest site voronoi diagram this reduces
    # the input complexity substantially and returns the same value
    points = convex.hull_points(obj)

    # if all of the points are on an n-sphere already the voronoi 
    # method will fail so we check a least squares fit before 
    # bothering to compute the voronoi diagram
    fit_C, fit_R, fit_E = fit_nsphere(points)
    radius_fit = ((points - fit_C)**2).sum(axis=1).max() ** .5
    if fit_E < 1e-3:
        log.debug('Points were on an n-sphere, returning fit')
        return fit_C, fit_R

    # calculate a furthest site voronoi diagram
    # this will fail if the points are ALL on the surface of
    # the n-sphere but hopefully the least squares check caught those cases
    voronoi = spatial.Voronoi(points, furthest_site=True, qhull_options='QbB Pp')

    # find the maximum radius^2 point for each of the voronoi vertices
    # this is worst case quite expensive, but we have used quick convex
    # hull methods to reduce n for this operation
    # we are doing comparisons on the radius^2 value so as to only do a sqrt once
    r2 = np.array([((points-v)**2).sum(axis=1).max() for v in voronoi.vertices])
    r2_idx = r2.argmin()
    radius_v = np.sqrt(r2[r2_idx])
    center_v = voronoi.vertices[r2_idx]

    if radius_v > radius_fit:
        return fit_C, radius_fit
    return center_v, radius_v

def fit_nsphere(points, prior=None):
    '''
    Fit an n-sphere to a set of points using least squares. 
    
    Arguments
    ---------
    points: (n,d) set of points
    prior:  (d,) float, best guess for center

    Returns
    ---------
    center: (d), location of center
    radius: float, mean radius across circle
    error:  float, peak to peak value of deviation from mean radius
    '''
    
    def residuals(center):
        radii_sq  = ((points-center)**2).sum(axis=1)
        residuals = radii_sq - radii_sq.mean()
        return residuals

    if prior is None: center_guess = np.mean(points, axis=0)
    else:             center_guess = prior

    center_result, return_code = leastsq(residuals, center_guess, gtol=1e-8)
    if not (return_code in [1,2,3,4]):
        raise ValueError('Least square fit failed!')

    radii  = np.linalg.norm(points-center_result, axis=1)
    radius = radii.mean()
    error  = radii.ptp()
    return center_result, radius, error



from .generic import MeshScript
from distutils.spawn import find_executable

_scad_executable = find_executable('openscad')
exists = _scad_executable is not None

def interface_scad(meshes, script):
    '''
    A way to interface with openSCAD which is itself an interface
    to the CGAL CSG bindings. 
    CGAL is very stable if difficult to install/use, so this function provides a 
    tempfile- happy solution for getting the basic CGAL CSG functionality. 

    Arguments
    ---------
    meshes: list of Trimesh objects
    script: string of the script to send to scad. 
            Trimesh objects can be referenced in the script as
            $mesh_0, $mesh_1, etc. 
    '''
    if not exists:
        raise ValueError('No SCAD available!')
    with MeshScript(meshes = meshes, script = script) as scad:
        result = scad.run(_scad_executable + ' $script -o $mesh_post')
    return result
    
def boolean(meshes, operation='difference'):
    '''
    Run an operation on a set of meshes
    '''
    script = operation + '(){'
    for i in range(len(meshes)):
        script += 'import(\"$mesh_' + str(i) + '\");'
    script += '}'
    return interface_scad(meshes, script)






from pkg_resources import resource_string

def get_resource(name):
    result = resource_string('trimesh',
                             'resources/' + name)
    if hasattr(result, 'decode'):
        return result.decode('utf-8')
    return result



import numpy as np
import pprint

from . import util
from . import creation

from .base      import Trimesh
from .constants import log
   
class _Primitive(Trimesh):
    '''
    Geometric _Primitives which are a subclass of Trimesh.
    Mesh is generated lazily when vertices or faces are requested.
    '''
    def __init__(self, *args, **kwargs):
        super(_Primitive, self).__init__(*args, **kwargs)
        self._data.clear()
        self._validate = False
        
    @property
    def faces(self):
        stored = self._cache['faces']
        if util.is_shape(stored, (-1,3)):
            return stored
        self._create_mesh()
        return self._cache['faces']

    @faces.setter
    def faces(self, values):
        log.warning('Primitive faces are immutable! Not setting!')

    @property
    def vertices(self):
        stored = self._cache['vertices']
        if util.is_shape(stored, (-1,3)):
            return stored

        self._create_mesh()
        return self._cache['vertices']

    @vertices.setter
    def vertices(self, values):
        if values is not None:
            log.warning('Primitive vertices are immutable! Not setting!')

    @property
    def face_normals(self):
        stored = self._cache['face_normals']
        if util.is_shape(stored, (-1,3)):
            return stored
        self._create_mesh()
        return self._cache['face_normals']
    
    @face_normals.setter
    def face_normals(self, values):
        if values is not None:
            log.warning('Primitive face normals are immutable! Not setting!')

    def to_mesh(self):
        '''
        Return a copy of the Primitive object as a trimesh.
        '''
        result = Trimesh(vertices = self.vertices.copy(),
                         faces    = self.faces.copy(),
                         face_normals = self.face_normals.copy())
        return result
            
    def _create_mesh(self):
        raise ValueError('Primitive doesn\'t define mesh creation!')

        
class _PrimitiveAttributes(object):
    def __init__(self, data, defaults, parent, kwargs):
        self._data     = data
        self._defaults = defaults
        self._data.update(defaults)
        self._mutable = True
        for key, value in kwargs.items():
            if key in defaults:
                self._data[key] = util.convert_like(value, defaults[key])
        # if configured as immutable, apply setting afer instantiation values are set
        if 'mutable' in kwargs:
            self._mutable = bool(kwargs['mutable'])
            
        self.__doc__ = (('Store the attributes of a {prim_name} object.\n\n' +
                        'When these values are changed, the mesh geometry will \n' +
                        'automatically be updated to reflect the new values.\n\n' +
                        'Available properties and their default values are:\n {prim_defaults}\n\n' +
                        'Example\n---------------\n' +
                        'p = trimesh.primitives.{prim_name}()\n' +
                        'p.primitive.radius = 10\n\n').format(prim_name = parent.__class__.__name__,
                                                              prim_defaults = pprint.pformat(defaults, width = -1)[1:-1]))

    def __getattr__(self, key):
        if '_' in key:
            return super(_PrimitiveAttributes, self).__getattr__(key)
        elif key in self._defaults:
            return util.convert_like(self._data[key], self._defaults[key])
        return super(_PrimitiveAttributes, self).__getattr__(key)
            
    def __setattr__(self, key, value):
        if '_' in key:
            return super(_PrimitiveAttributes, self).__setattr__(key, value) 
        elif key in self._defaults:
            if self._mutable:
                self._data[key] = util.convert_like(value,  self._defaults[key])
            else:
                raise ValueError('Primitive is configured as immutable! Cannot set attribute!')
        else:
            keys = list(self._defaults.keys())
            raise ValueError('Only default attributes {} can be set!'.format(keys))
            
    def __dir__(self):
        result = sorted(dir(type(self)) + 
                        list(self._defaults.keys()))
        return result
            
class Cylinder(_Primitive):
    def __init__(self, *args, **kwargs):
        '''
        Create a Cylinder Primitive, a subclass of Trimesh.

        Arguments
        ----------
        radius: float, radius of cylinder
        height: float, height of cylinder
        transform: (4,4) float, transformation matrix
        sections: int, number of facets in circle
        '''
        super(Cylinder, self).__init__(*args, **kwargs)
        
        defaults = {'height'    : 1.0,
                    'radius'    : 1.0,
                    'transform' : np.eye(4),
                    'sections'  : 32}
        self.primitive = _PrimitiveAttributes(self._data, defaults, self, kwargs)

    def _create_mesh(self):
        log.info('Creating cylinder mesh with r=%f, h=%f and %d sections',
                 self.primitive.radius,
                 self.primitive.height,
                 self.primitive.sections)
                 
        mesh = creation.cylinder(radius   = self.primitive.radius,
                                 height   = self.primitive.height,
                                 sections = self.primitive.sections)
        mesh.apply_transform(self.primitive.transform)
        
        self._cache['vertices']     = mesh.vertices
        self._cache['faces']        = mesh.faces
        self._cache['face_normals'] = mesh.face_normals
        
        
class Sphere(_Primitive):
    def __init__(self, *args, **kwargs):
        '''
        Create a Sphere Primitive, a subclass of Trimesh.

        Arguments
        ----------
        radius: float, radius of sphere
        center: (3,) float, center of sphere
        subdivisions: int, number of subdivisions for icosphere. Default is 3
        '''

        super(Sphere, self).__init__(*args, **kwargs)
        
        defaults = {'radius'        : 1.0,
                    'center'        : np.zeros(3, dtype=np.float64),
                    'subdivisions'  : 3}
                    
        self.primitive = _PrimitiveAttributes(self._data, defaults, self, kwargs)

    @property
    def bounds(self):
        # no docstring so will inherit Trimesh docstring
        # return exact bounds from primitive center and radius (rather than faces)
        # self.extents will also use this information
        bounds = np.array([self.primitive.center - self.primitive.radius,
                           self.primitive.center + self.primitive.radius])
        return bounds
        
    @property
    def bounding_box_oriented(self):
        # for a sphere the oriented bounding box is the same as the axis aligned 
        # bounding box, and a sphere is the absolute slowest case for the OBB calculation
        # as it is a convex surface with a ton of face normals that all need to be checked
        
        return self.bounding_box
		
    @util.cache_decorator
    def volume(self):
        '''
        Volume of the current sphere Primitive.

        Returns
        --------
        volume: float, volume of the sphere Primitive
        '''
        
        volume = (4.0*np.pi * (self.primitive.radius** 3)) / 3.0
        return volume
    
    def _create_mesh(self):
        unit = creation.icosphere(subdivisions=self.primitive.subdivisions)
        unit.vertices *= self.primitive.radius
        unit.vertices += self.primitive.center
        
        self._cache['vertices']     = unit.vertices
        self._cache['faces']        = unit.faces
        self._cache['face_normals'] = unit.face_normals

class Box(_Primitive):    
    def __init__(self, *args, **kwargs):
        '''
        Create a Box Primitive, a subclass of Trimesh

        Arguments
        ----------
        extents:   (3,)  float, size of box
        transform: (4,4) float, transformation matrix for box center
        '''
        super(Box, self).__init__(*args, **kwargs)
     
        defaults = {'transform' : np.eye(4),
                    'extents'   : np.ones(3)}
        self.primitive = _PrimitiveAttributes(self._data, defaults, self, kwargs)

    @property
    def is_oriented(self):
        '''
        Returns whether or not the current box is rotated at all.
        '''
        if util.is_shape(self.primitive.transform, (4,4)):
            return not np.allclose(self.primitive.transform[0:3,0:3], np.eye(3))
        else:
            return False

    @util.cache_decorator
    def volume(self):
        '''
        Volume of the box Primitive.

        Returns
        --------
        volume: float, volume of box
        '''
        volume = float(np.product(self.primitive.extents))
        return volume
        
    def _create_mesh(self):
        log.debug('Creating mesh for box Primitive')
        box = creation.box(extents   = self.primitive.extents,
                           transform = self.primitive.transform)
                              
        self._cache['vertices']     = box.vertices
        self._cache['faces']        = box.faces
        self._cache['face_normals'] = box.face_normals

class Extrusion(_Primitive):
    def __init__(self, *args, **kwargs):
        '''
        Create an Extrusion Primitive, a subclass of Trimesh

        Arguments
        ----------
        polygon:   shapely.geometry.Polygon, polygon to extrude
        transform: (4,4) float, transform to apply after extrusion
        height:    float, height to extrude polygon by
        '''
        super(Extrusion, self).__init__(*args, **kwargs)

        defaults = {'polygon'   : None,
                    'transform' : np.eye(4),
                    'height'    : 1.0}
                    
        self.primitive = _PrimitiveAttributes(self._data, defaults, self, kwargs)

    @property
    def extrude_direction(self):
        '''
        Based on the extrudes transform, what is the vector along
        which the polygon will be extruded
        
        Returns
        ---------
        direction: (3,) float vector. If self.primitive.transform is an 
                   identity matrix this will be [0.0, 0.0, 1.0]
        '''
        direction = np.dot(self.primitive.transform[:3,:3], 
                           [0.0,0.0,1.0])
        return direction
    
    def slide(self, distance):
        '''
        Alter the transform of the current extrusion to slide it along its
        extrude_direction vector
        
        Arguments
        -----------
        distance: float, distance along self.extrude_direction to move
        '''
        distance = float(distance)
        translation = np.eye(4)
        translation[2,3] = distance
        new_transform = np.dot(self.primitive.transform.copy(),
                               translation.copy())
        self.primitive.transform = new_transform

    def _create_mesh(self):
        log.debug('Creating mesh for extrude Primitive')
        mesh = creation.extrude_polygon(self.primitive.polygon[0],
                                        self.primitive.height)
        mesh.apply_transform(self.primitive.transform)
        self._cache['vertices']     = mesh.vertices
        self._cache['faces']        = mesh.faces
        self._cache['face_normals'] = mesh.face_normals



import numpy as np
import json
from string import Template

from ..arc        import arc_center
from ...resources import get_resource
from ...util      import three_dimensionalize
from ...constants import log
from ...constants import res_path as res

_templates_dxf = {k:Template(v) for k,v in json.loads(get_resource('dxf.json.template')).items()}

def export_path(path, file_type, file_obj=None):
    '''
    Export a Path object to a file- like object, or to a filename

    Arguments
    ---------
    file_obj:  a filename string or a file-like object
    file_type: str representing file type (eg: 'svg')
    process:   boolean flag, whether to process the mesh on load

    Returns
    ---------
    mesh: a single Trimesh object, or a list of Trimesh objects, 
          depending on the file format. 
    
    '''
    if ((not hasattr(file_obj, 'read')) and 
        (not file_obj is None)):
        file_type = (str(file_obj).split('.')[-1]).lower()
        file_obj  = open(file_obj, 'wb')
    export = _path_exporters[file_type](path)
    return _write_export(export, file_obj)

def export_dict(path):
    export_entities = [e.to_dict() for e in path.entities]
    export_object   = {'entities' : export_entities,
                       'vertices' : path.vertices.tolist()}
    return export_object

def export_svg(drawing):
    '''
    Will turn a path drawing into an SVG path string. 

    'holes' will be in reverse order, so they can be rendered as holes by
    rendering libraries
    '''
    points = drawing.vertices.view(np.ndarray)
    def circle_to_svgpath(center, radius, reverse):
        radius_str = format(radius, res.export)
        path_str  = '  M' + format(center[0]-radius, res.export) + ',' 
        path_str += format(center[1], res.export)       
        path_str += 'a' + radius_str + ',' + radius_str  
        path_str += ',0,1,' + str(int(reverse)) + ','
        path_str += format(2*radius, res.export) +  ',0'
        path_str += 'a' + radius_str + ',' + radius_str
        path_str += ',0,1,' + str(int(reverse)) + ','
        path_str += format(-2*radius, res.export) + ',0Z  '
        return path_str
    def svg_arc(arc, reverse):
        '''
        arc string: (rx ry x-axis-rotation large-arc-flag sweep-flag x y)+
        large-arc-flag: greater than 180 degrees
        sweep flag: direction (cw/ccw)
        '''
        vertices = points[arc.points[::((reverse*-2) + 1)]]        
        vertex_start, vertex_mid, vertex_end = vertices
        center_info = arc_center(vertices)
        C, R, N, angle = (center_info['center'],
                          center_info['radius'],
                          center_info['normal'],
                          center_info['span'])
        if arc.closed: return circle_to_svgpath(C, R, reverse)
        large_flag = str(int(angle > np.pi))
        sweep_flag = str(int(np.cross(vertex_mid-vertex_start, 
                                      vertex_end-vertex_start) > 0))
        R_ex = format(R, res.export)
        x_ex = format(vertex_end[0],res.export)
        y_ex = format(vertex_end [1],res.export)
        arc_str  = 'A' + R_ex + ',' + R_ex + ' 0 ' 
        arc_str += large_flag + ',' + sweep_flag + ' '
        arc_str += x_ex + ',' + y_ex
        return arc_str
    def svg_line(line, reverse):
        vertex_end = points[line.points[-(not reverse)]]
        x_ex = format(vertex_end[0], res.export) 
        y_ex = format(vertex_end[1], res.export) 
        line_str = 'L' + x_ex + ',' + y_ex
        return line_str
    def svg_moveto(vertex_id):
        x_ex = format(points[vertex_id][0], res.export) 
        y_ex = format(points[vertex_id][1], res.export) 
        move_str = 'M' + x_ex + ',' + y_ex
        return move_str
    def convert_path(path, reverse=False):
        path     = path[::(reverse*-2) + 1]
        path_str = svg_moveto(drawing.entities[path[0]].end_points[-reverse])
        for i, entity_id in enumerate(path):
            entity = drawing.entities[entity_id]
            e_type = entity.__class__.__name__
            try: 
                path_str += converters[e_type](entity, reverse)
            except KeyError:
                log.warn('%s entity not available for export!', e_type)
        path_str += 'Z'
        return path_str

    converters = {'Line'  : svg_line,
                  'Arc'   : svg_arc}
    path_str = ''
    for path_index, path in enumerate(drawing.paths):
        reverse   = not (path_index in drawing.root)
        path_str += convert_path(path, reverse)
    return path_str

def export_dxf(path):
    '''
    Export a 2D path object to a DXF file
    
    Arguments
    ----------
    path: trimesh.path.path.Path2D
    
    Returns
    ----------
    export: str, path formatted as a DXF file
    '''
    
    def format_points(points, increment = True):
        points = np.asanyarray(points)
        three = three_dimensionalize(points, return_2D =False)
        if increment:
            group = np.tile(np.arange(len(three)).reshape((-1,1)), (1,3)) 
        else:
            group = np.zeros((len(three), 3), dtype=np.int)
        group += [10, 20, 30]
        interleaved = np.dstack((group.astype(str),
                                 three.astype(str))).reshape(-1)
        packed = '\n'.join(interleaved)
        return packed
    def entity_color_layer(entity):
        color, layer = 0, 0
        if hasattr(entity, 'color'): color = int(entity.color)
        if hasattr(entity, 'layer'): layer = int(entity.layer)
        return color, layer
    def convert_line(line, vertices):
        points = line.discrete(vertices)
        color, layer = entity_color_layer(line)
        is_poly = len(points) > 2
        line_type = ['LINE', 'LWPOLYLINE'][int(is_poly)]
        result = templates['line'].substitute({'TYPE'         : line_type,
                                               'POINTS'       : format_points(points,
                                                                              increment = not is_poly),
                                               'NAME'         : str(id(line))[:16],
                                               'LAYER_NUMBER' : layer,
                                               'COLOR_NUMBER' : color})
        return result
    def convert_arc(arc, vertices):
        info = arc.center(vertices)
        color, layer = entity_color_layer(arc)
        angles = np.degrees(info['angles'])
        arc_type = ['ARC', 'CIRCLE'][int(arc.closed)]
        result = templates['arc'].substitute({'TYPE' : arc_type,
                                              'CENTER_POINT' : format_points([info['center']]),
                                              'ANGLE_MIN'    : angles[0],
                                              'ANGLE_MAX'    : angles[1],
                                              'RADIUS'       : info['radius'],
                                              'LAYER_NUMBER' : layer,
                                              'COLOR_NUMBER' : color})
        return result

    def convert_generic(entity, vertices):
        log.warning('No converter available! Exporting as polyline!')
        return convert_line(entity, vertices)

    templates = _templates_dxf
    np.set_printoptions(precision=12)
    conversions = {'Line'    : convert_line,
                   'Arc'     : convert_arc,
                   'Bezier'  : convert_generic,
                   'BSpline' : convert_generic}
    entities_str = ''
    for e in path.entities:
        name = type(e).__name__
        if name in conversions:
            entities_str += conversions[name](e, path.vertices)
        else:
            log.warning('Entity type %s not exported!', name)

    header = templates['header'].substitute({'BOUNDS_MIN': format_points([path.bounds[0]]),
                                             'BOUNDS_MAX': format_points([path.bounds[1]]),
                                             'UNITS_CODE': '1'})
    entities = templates['entities'].substitute({'ENTITIES' : entities_str})
    footer   = templates['footer'].substitute()
    export   = '\n'.join([header, entities, footer])
    return export

def _write_export(export, file_obj=None):
    '''
    Write a string to a file.
    If file_obj isn't specified, return the string

    Arguments
    ---------
    export: a string of the export data
    file_obj: a file-like object or a filename
    '''

    if file_obj is None:             
        return export
    elif hasattr(file_obj, 'write'): 
        out_file = file_obj
    else: 
        out_file = open(file_obj, 'wb')
    try:
        out_file.write(export)
    except TypeError:
        out_file.write(export.encode('utf-8'))
    out_file.close()
    return export

_path_exporters = {'dxf'  : export_dxf,
                   'svg'  : export_svg,
                   'dict' : export_dict}



import numpy as np

from ..util            import Cache, unitize
from ..grouping        import unique_rows
from ..intersections   import plane_lines
from .ray_triangle_cpu import rays_triangles_id

class RayMeshIntersector:
    '''
    An object to query a mesh for ray intersections. 
    Precomputes an r-tree for each triangle on the mesh.
    '''
    def __init__(self, mesh):
        self.mesh   = mesh
        self._cache = Cache(self.mesh.md5)

    @property
    def tree(self):
        if 'tree' in self._cache:
            return self._cache.get('tree')
        else:
            return self._cache.set('tree',
                                   self.mesh.triangles_tree())

    def intersects_id(self, rays, return_any=False):
        '''
        Find the indexes of triangles the rays intersect

        Arguments
        ---------
        rays: (n, 2, 3) array of ray origins and directions

        Returns
        ---------        
        hits: (n) sequence of triangle indexes which hit the ray
        '''
        rays       = np.array(rays, dtype=np.float)
        candidates = ray_triangle_candidates(rays = rays, 
                                             tree = self.tree)
        hits  = rays_triangles_id(triangles      = self.mesh.triangles, 
                                  rays           = rays, 
                                  ray_candidates = candidates,
                                  return_any     = return_any)
        return hits
            
    def intersects_location(self, rays, return_id=False):
        '''
        Return unique cartesian locations where rays hit the mesh.
        If you are counting the number of hits a ray had, this method 
        should be used as if only the triangle index is used on- edge hits
        will be counted twice. 

        Arguments
        ---------
        rays: (n, 2, 3) array of ray origins and directions
        return_id: boolean flag, if True return triangle indexes
        Returns
        ---------
        locations: (n) sequence of (m,3) intersection points
        hits:      (n) list of face ids 
        '''
        rays      = np.array(rays, dtype=np.float)
        hits      = self.intersects_id(rays)
        locations = ray_triangle_locations(triangles     = self.mesh.triangles,
                                                  rays          = rays,
                                                  intersections = hits,
                                                  tri_normals   = self.mesh.face_normals)
        if return_id:
            return locations, hits
        return locations

    def intersects_any_triangle(self, rays):
        '''
        Find out whether the rays in question hit *any* triangle on the mesh.

        Arguments
        ---------
        rays: (n, 2, 3) array of ray origins and directions

        Returns
        ---------
        hits_any: (n) boolean array of whether or not each ray hit any triangle
        '''
        hits     = self.intersects_id(rays)
        hits_any = np.array([len(i) > 0 for i in hits])
        return hits_any

    def intersects_any(self, rays):
        '''
        Find out whether *any* ray hit *any* triangle on the mesh.
        Equivilant to but signifigantly faster than (due to early exit):
            intersects_any_triangle(rays).any()

        Arguments
        ---------
        rays: (n, 2, 3) array of ray origins and directions

        Returns
        ---------
        hit: boolean, whether any ray hit any triangle on the mesh
        '''
        hit = self.intersects_id(rays, return_any=True)
        return hit

def ray_triangle_candidates(rays, tree):
    '''
    Do broad- phase search for triangles that the rays
    may intersect. 

    Does this by creating a bounding box for the ray as it 
    passes through the volume occupied by the tree
    '''
    ray_bounding   = ray_bounds(rays, tree.bounds)
    ray_candidates = [None] * len(rays)
    for ray_index, bounds in enumerate(ray_bounding):
        ray_candidates[ray_index] = list(tree.intersection(bounds))
    return ray_candidates

def ray_bounds(rays, bounds, buffer_dist = 1e-5):
    '''
    Given a set of rays and a bounding box for the volume of interest
    where the rays will be passing through, find the bounding boxes 
    of the rays as they pass through the volume. 
    
    Arguments
    ---------
    rays: (n,2,3) array of ray origins and directions
    bounds: (2,3) bounding box (min, max)
    buffer_dist: float, distance to pad zero width bounding boxes

    Returns
    ---------
    ray_bounding: (n) set of AABB of rays passing through volume
    '''
    # separate out the (n, 2, 3) rays array into (n, 3) 
    # origin/direction arrays
    ray_ori = rays[:,0,:]
    ray_dir = unitize(rays[:,1,:])

    # bounding box we are testing against
    bounds  = np.array(bounds)

    # find the primary axis of the vector
    axis       = np.abs(ray_dir).argmax(axis=1)
    axis_bound = bounds.reshape((2,-1)).T[axis]
    axis_ori   = np.array([ray_ori[i][a] for i, a in enumerate(axis)]).reshape((-1,1))
    axis_dir   = np.array([ray_dir[i][a] for i, a in enumerate(axis)]).reshape((-1,1))

    # parametric equation of a line
    # point = direction*t + origin
    # p = dt + o
    # t = (p-o)/d
    t = (axis_bound - axis_ori) / axis_dir

    # prevent the bounding box from including triangles
    # behind the ray origin
    t[t < buffer_dist] = buffer_dist

    # the value of t for both the upper and lower bounds
    t_a = t[:,0].reshape((-1,1))
    t_b = t[:,1].reshape((-1,1))

    # the cartesion point for where the line hits the plane defined by
    # axis
    on_a = (ray_dir * t_a) + ray_ori
    on_b = (ray_dir * t_b) + ray_ori

    on_plane = np.column_stack((on_a, on_b)).reshape((-1,2,ray_dir.shape[1]))
    
    ray_bounding = np.hstack((on_plane.min(axis=1), on_plane.max(axis=1)))
    # pad the bounding box by TOL_BUFFER
    # not sure if this is necessary, but if the ray is  axis aligned
    # this function will otherwise return zero volume bounding boxes
    # which may or may not screw up the r-tree intersection queries
    ray_bounding += np.array([-1,-1,-1,1,1,1]) * buffer_dist

    return ray_bounding

def ray_triangle_locations(triangles, 
                           rays, 
                           intersections, 
                           tri_normals):
    '''
    Given a set of triangles, rays, and intersections between the two,
    find the cartesian locations of the intersections points. 

    Arguments
    ----------
    triangles:     (n, 3, 3) set of triangle vertices
    rays:          (m, 2, 3) set of ray origins/ray direction pairs
    intersections: (m) sequence of intersection indidices which triangles
                    each ray hits. 

    Returns
    ----------
    locations: (m) sequence of (p,3) cartesian points
    '''
    
    ray_origin   = rays[:,0,:]
    ray_vector   = rays[:,1,:]
    ray_segments = np.array([ray_origin,
                             ray_origin + ray_vector])
    locations = [[]] * len(rays)

    for r, tri_group in enumerate(intersections):
        group_locations = np.zeros((len(tri_group), 3))
        valid = np.zeros(len(tri_group), dtype=np.bool)
        for i, tri_index in enumerate(tri_group):
            origin  = triangles[tri_index][0]
            normal  = tri_normals[tri_index]
            segment = ray_segments[:,r,:].reshape((2,-1,3))
            point, ok = plane_lines(plane_origin  = origin,
                                   plane_normal  = normal,
                                   endpoints     = segment,
                                   line_segments = False)
            if ok:
                valid[i] = True
                group_locations[i] = point
        group_locations = group_locations[valid]
        unique = unique_rows(group_locations)[0]
        locations[r] = group_locations[unique]
    return np.array(locations)

def contains_points(mesh, points):
    '''
    Check if a mesh contains a set of points, using ray tests.

    If the point is on the surface of the mesh, behavior is undefined.

    Arguments
    ---------
    mesh: Trimesh object
    points: (n,3) points in space

    Returns
    ---------
    contains: (n) boolean array, whether point is inside mesh or not
    '''
    points = np.asanyarray(points)
    vector = unitize([0,0,1])
    rays = np.column_stack((points, 
                            np.tile(vector,(len(points),1)))).reshape((-1,2,3))
    hits = mesh.ray.intersects_location(rays)
    hits_count = np.array([len(i) for i in hits])
    contains = np.mod(hits_count, 2) == 1
  
    return contains



from .constants import log

# conversions to inches
_TO_INCHES = {'microinches' : 1.0 / 1000.0, 
              'mils'        : 1.0 / 1000.0,
              'inches'      : 1.00,
              'feet'        : 12.0,
              'yards'       : 36.0,
              'miles'       : 63360,
              'angstroms'   : 1.0 / 2.54e8,
              'nanometers'  : 1.0 / 2.54e7,
              'microns'     : 1.0 / 2.54e4,
              'millimeters' : 1.0 / 2.54e1, 
              'centimeters' : 1.0 / 2.54e0,
              'meters'      : 1.0 / 2.54e-2,
              'kilometers'  : 1.0 / 2.54e-5,
              'decimeters'  : 1.0 / 2.54e-1,
              'decameters'  : 1.0 / 2.54e-3,
              'hectometers' : 1.0 / 2.54e-4,
              'gigameters'  : 1.0 / 2.54e-11,
              'AU'          : 5889679948818.897,
              'light years' : 3.72461748e17,
              'parsecs'     : 1.21483369e18}

# if a unit is known by other symbols, include them here
_synonyms = {'millimeters' : ['mm'],
             'inches'      : ['in'],
             'meters'      : ['m']}
             
for key, new_keys in _synonyms.items():
    _value = _TO_INCHES[key]
    for new_key in new_keys:
        _TO_INCHES[new_key] = _value

def unit_conversion(current, desired):
    '''
    Calculate the conversion from one set of units to another.

    Arguments
    ---------
    current: str, unit system values are in now (eg 'millimeters')
    desired: str, unit system we'd like values in (eg 'inches')

    Returns
    ---------
    conversion: float, number to multiply by to put values into desired units
    '''
    conversion = _TO_INCHES[current] / _TO_INCHES[desired]
    return conversion

def validate(units):
    '''
    Check whether a string represents the name of a valid unit
    
    Returns
    ---------
    valid: bool, is units string a valid unit or not
    '''
    valid = str(units) in _TO_INCHES
    return valid
    
def unit_guess(scale):
    '''
    Wild ass guess for the units of a drawing or model, based on the scale.
    '''
    if scale > 100.0: 
        return 'millimeters'
    else:      
        return 'inches'

def _set_units(obj, desired, guess):
    '''
    Given an object that has units and vertices attributes convert units.

    Arguments
    ---------
    obj:     object with units and vertices (eg Path or Trimesh)
    desired: units desired (eg 'inches')
    guess:   boolean, whether we are allowed to guess the units of the document
             if they are not specified. 
    '''
    desired = str(desired)
    if not validate(desired):
        raise ValueError(desired + ' are not a valid unit!')

    if obj.units is None:
        if guess:
            obj.units = unit_guess(obj.scale)
            log.warn('No units specified, guessing current units are %s', obj.units)
        else:
            raise ValueError('No units specified and not allowed to guess!')
    log.info('Converting units from %s to %s', obj.units, desired)
    conversion    = unit_conversion(obj.units, desired)
    obj.vertices *= conversion
    obj.units     = desired





from time import time as time_function
from logging import getLogger   as _getLogger
from logging import NullHandler as _NullHandler

### numerical tolerances for meshes
class NumericalToleranceMesh(object):
    '''
    tol.zero: consider floating point numbers less than this zero
    tol.merge: when merging vertices, consider vertices closer than this
               to be the same vertex. Here we use the same value (1e-8)
               as SolidWorks uses, according to their documentation.
    tol.planar: the maximum distance from a plane a point can be and
                still be considered to be on the plane
    tol.facet_rsq: the minimum radius squared that an arc drawn from the 
                   center of a face to the center of an adjacent face can
                   be to consider the two faces coplanar. This method is more
                   robust than considering just normal angles as it is tolerant
                   of numerical error on very small faces. 
    '''
    def __init__(self, **kwargs):
        self.zero      = 1e-12
        self.merge     = 1e-8
        self.planar    = 1e-5
        self.facet_rsq = 1e8
        self.fit       = 1e-2
        self.id_len    = 6
        self.__dict__.update(kwargs)


class NumericalResolutionMesh(object):
    '''
    res.mesh: when meshing solids, what distance to use
    '''
    def __init__(self, **kwargs):
        self.mesh = 5e-3

        self.__dict__.update(kwargs)

tol = NumericalToleranceMesh()
res = NumericalResolutionMesh()

### numerical tolerances for paths
class NumericalTolerancePath(object):
    '''
    tol.zero: consider floating point numbers less than this zero
    tol.merge: when merging vertices, consider vertices closer than this
               to be the same vertex
    tol.planar: the maximum distance from a plane a point can be and
                still be considered to be on the plane
    tol.seg_frac: when simplifying line segments what percentage of the drawing
                  scale can a segment be and have a curve fitted
    tol.seg_angle: when simplifying line segments to arcs, what angle
                   can a segment span to be acceptable.
    tol.aspect_frac: when simplifying line segments to closed arcs (circles)
                     what percentage can the aspect ratio differfrom 1:1
                     before escaping the fit early
    tol.radius_frac: when simplifying line segments to arcs, what percentage 
                     of the fit radius can vertices deviate to be acceptable
    tol.radius_min: when simplifying line segments to arcs, what is the minimum 
                    radius multiplied by document scale for an acceptable fit
    tol.radius_max: when simplifying line segments to arcs, what is the maximum
                    radius multiplied by document scale for an acceptable fit 
    tol.tangent: when simplifying line segments to curves, what is the maximum
                 angle the end sections can deviate from tangent that is acceptable.   
    '''
    def __init__(self, **kwargs):
        # default values
        self.zero        = 1e-12
        self.merge       = 1e-5
        self.planar      = 1e-5
        self.buffer      = .05
        self.seg_frac    = .125
        self.seg_angle   = .8
        self.aspect_frac = .1
        self.radius_frac = 1e-2
        self.radius_min  = 1e-4
        self.radius_max  = 50
        self.tangent     = .017

        self.__dict__.update(kwargs)

class NumericalResolutionPath(object):
    '''
    res.seg_frac: when discretizing curves, what percentage of the drawing
                  scale should we aim to make a single segment
    res.seg_angle: when discretizing curves, what angle should a section span
    res.max_sections: when discretizing splines, what is the maximum number 
                      of segments per control point
    res.min_sections: when discretizing splines, what is the minimum number
                      of segments per control point
    res.export: format string to use when exporting floating point vertices
    '''
    def __init__(self, **kwargs):
        self.seg_frac     = .05
        self.seg_angle    = .08
        self.max_sections = 10
        self.min_sections = 5
        self.export       = '.5f'

tol_path = NumericalTolerancePath()
res_path = NumericalResolutionPath()

### logging
log = _getLogger('trimesh')
log.addHandler(_NullHandler())
def _log_time(method):
    def timed(*args, **kwargs):
        tic    = time_function()
        result = method(*args, **kwargs)
        log.debug('%s executed in %.4f seconds.',
                   method.__name__,
                   time_function()-tic)
        return result
    timed.__name__ = method.__name__
    timed.__doc__  = method.__doc__
    return timed

### exceptions
class MeshError(Exception):
    pass
class TransformError(Exception): 
    pass



'''
Narrow phase ray- triangle intersection
'''
import numpy as np

from ..constants import tol
from ..util      import diagonal_dot

def rays_triangles_id(triangles,
                      rays, 
                      ray_candidates = None,
                      return_any     = False):
    '''
    Intersect a set of rays and triangles. 

    Arguments
    ---------
    triangles:      (n, 3, 3) float array of triangle vertices
    rays:           (m, 2, 3) float array of ray start, ray directions
    ray_candidates: (m, *) int array of which triangles are candidates
                    for the ray. 
    return_any:     bool, exit loop early if any ray hits any triangle
                    and change output of function to bool

    Returns
    ---------
    if return_any:
        hit:           bool, whether the set of rays hit any triangle
    else:
        intersections: (m) sequence of triangle indexes hit by rays
    '''

    # default set of candidate triangles to be queried 
    # is every triangle. this is very slow
    candidates = np.ones(len(triangles), dtype=np.bool)
    hits       = [None] * len(rays)

    for ray_index, ray in enumerate(rays):
        if not (ray_candidates is None):
            candidates = ray_candidates[ray_index]
        # query the triangle candidates
        hit = ray_triangles(triangles[candidates], *ray)
        if return_any:
            if hit.any(): return True
        else: 
            hits[ray_index] = np.array(candidates)[hit]

    if return_any: return False
    return np.array(hits)

def ray_triangles(triangles, 
                  ray_origin, 
                  ray_direction):
    '''
    Intersection of multiple triangles and a single ray.

    Moller-Trumbore intersection algorithm.
    '''
    candidates = np.ones(len(triangles), dtype=np.bool)

    # edge vectors and vertex locations in (n,3) format
    vert0 = triangles[:,0,:]
    vert1 = triangles[:,1,:]
    vert2 = triangles[:,2,:]
    edge0 = vert1 - vert0
    edge1 = vert2 - vert0

    #P is a vector perpendicular to the ray direction and one
    # triangle edge. 
    P = np.cross(ray_direction, edge1)

    #if determinant is near zero, ray lies in plane of triangle
    det = diagonal_dot(edge0, P)
    candidates[np.abs(det) < tol.zero] = False

    if not candidates.any(): 
        return candidates
    # remove previously calculated terms which are no longer candidates
    inv_det = 1.0 / det[candidates]
    T       = ray_origin - vert0[candidates]
    u       = diagonal_dot(T, P[candidates]) * inv_det

    new_candidates = np.logical_not(np.logical_or(u < -tol.zero,
                                                  u > (1+tol.zero)))
    candidates[candidates] = new_candidates
    if not candidates.any(): 
        return candidates    
    inv_det = inv_det[new_candidates]
    T       = T[new_candidates]
    u       = u[new_candidates]

    Q = np.cross(T, edge0[candidates])
    v = np.dot(ray_direction, Q.T) * inv_det

    new_candidates = np.logical_not(np.logical_or((v     <  -tol.zero),
                                                  (u + v > (1+tol.zero))))

    candidates[candidates] = new_candidates
    if not candidates.any(): 
        return candidates

    Q = Q[new_candidates]
    inv_det = inv_det[new_candidates]
    
    t = diagonal_dot(edge1[candidates], Q) * inv_det
    candidates[candidates] = t > tol.zero

    return candidates



import numpy as np 

from collections import deque

from ..constants import log, time_function
from ..constants import tol_path as tol

from .polygons  import polygons_obb, transform_polygon

class RectangleBin:
    '''
    2D BSP tree node. 
    http://www.blackpawn.com/texts/lightmaps/
    '''
    def __init__(self, bounds=None, size=None):
        self.child  = [None] * 2

        #bounds: (minx, miny, maxx, maxy)
        self.bounds   = bounds
        self.occupied = False
        
        if size != None:  
            self.bounds = np.append([0,0], size)
        
    def insert(self, rectangle_size):
        for child in self.child:
            if child != None: 
                attempt = child.insert(rectangle_size)
                if attempt: return attempt
                
        if self.occupied:
            return None

        # compare the bin size to the insertion candidate size
        size_test = bounds_to_size(self.bounds) - rectangle_size

        # this means the inserted rectangle is too big for the cell
        if np.any(size_test < -tol.zero):
            return None
        
        # since the cell is big enough for the current rectangle, either it
        # is going to be inserted here, or the cell is going to be split
        # either way, the cell is now occupied. 
        self.occupied = True    
        
        # this means the inserted rectangle fits perfectly
        # since we already checked to see if it was negative, no abs is needed
        if np.all(size_test < tol.zero): 
            return self.bounds[0:2]
        
        # since the rectangle fits but the empty space is too big, 
        # we need to create some children to insert into
        # first, we decide which way to split
        vertical      = size_test[0] > size_test[1]
        length        = rectangle_size[int(not vertical)]
        child_bounds  = self.split(length, vertical)

        self.child[0] = RectangleBin(bounds=child_bounds[0])
        self.child[1] = RectangleBin(bounds=child_bounds[1])
        
        return self.child[0].insert(rectangle_size)
   
    def split(self, length=None, vertical=True):
        '''
        returns two bounding boxes representing the current
        bounds split into two smaller boxes
        '''
        #also know as [minx, miny, maxx, maxy]
        [left, bottom, right, top] = self.bounds
        if vertical:
            box = [[left,        bottom, left+length, top],
                   [left+length, bottom, right,       top]]
        else: 
            box = [[left, bottom,          right, bottom + length],
                   [left, bottom + length, right, top]]
        return box

def bounds_to_size(bounds):
    return np.diff(np.reshape(bounds, (2,2)), axis=0)[0]    

def pack_rectangles(rectangles, sheet_size, shuffle=False):
    '''
    Pack smaller rectangles onto a larger rectangle, using a binary space partition tree.

    Parameters
    ----------
    rectangles: (n,2) array of (width, height) pairs representing the smaller rectangles to be packed.
    sheet_size: (2) array of (width, height) pair representing the sheet size the smaller rectangles will be packed onto.
    shuffle: boolean, whether or not to shuffle the insert order of the smaller rectangles, as the final packing density depends on the order of which rectangles are inserted onto the larger sheet. 
    density_escape: float, at what density should the loop exit early and return. A value of zero will return immediately after a single rectangle is placed, and a value of >= 1.0 will always traverse the entire list of smaller rectangles.
    '''
    offset    = np.zeros((len(rectangles), 2))
    inserted  = np.zeros( len(rectangles), dtype=np.bool)
    box_order = np.argsort(np.sum(rectangles**2, axis=1))[::-1]
    area      = 0.0
    density   = 0.0
    
    if shuffle: 
        shuffle_len = int(np.random.random()* len(rectangles)) - 1  
        box_order[0:shuffle_len] = np.random.permutation(box_order[0:shuffle_len])

    sheet = RectangleBin(size=sheet_size)
    for index in box_order: 
        insert_location = sheet.insert(rectangles[index])
        if insert_location != None:
            area           += np.prod(rectangles[index])
            offset[index]  += insert_location
            inserted[index] = True
    
    consumed_box = np.max((offset + rectangles)[inserted], axis=0)
    density      = area / np.product(consumed_box) 
            
    return density, offset[inserted], inserted, consumed_box
  
def pack_paths(paths, show=False):
    paths_full = deque()
    for path in paths:
        if 'quantity' in path.metadata:
            paths_full.extend([path.copy() for i in range(path.metadata['quantity'])])
        else:
            paths_full.append(path.copy())

    polygons   = [i.polygons_closed[i.root[0]] for i in paths_full]
    inserted, transforms = multipack(np.array(polygons))
    for path, transform in zip(paths_full, transforms):
        if show: path.plot_discrete(show=False, transform=transform)
    if show:
        import matplotlib.pyplot as plt
        plt.show()
    return paths_full

def multipack(polygons, 
              sheet_size     = None,
              iterations     = 50,
              density_escape = .985,
              buffer_dist    = 0.09,
              plot           = False,
              return_all     = False):
    '''
    Run multiple iterations of rectangle packing, by randomly permutating the insertion order

    If sheet size isn't specified, it creates a large sheet that can fit all of the polygons
    '''
    transforms_obb, rectangles = polygons_obb(polygons)
    rectangles                += 2.0*buffer_dist
    polygon_area               = np.array([p.area for p in polygons])


    for i, r in enumerate(rectangles):
        transforms_obb[i][0:2,2] += r*.5
    

    tic             = time_function()  
    overall_density = 0
    
    if sheet_size==None:
        max_dim    = np.max(rectangles, axis=0)
        sum_dim    = np.sum(rectangles, axis=0)
        sheet_size = [sum_dim[0], max_dim[1]*2]

    log.info('Packing %d polygons', len(polygons))
    for i in range(iterations):
        density, offset, inserted, sheet = pack_rectangles(rectangles, 
                                                           sheet_size = sheet_size, 
                                                           shuffle    = (i != 0))
        if density > overall_density:
            overall_density  = density
            overall_offset   = offset
            overall_inserted = inserted
            overall_sheet    = sheet
            if density > density_escape: break
            
    toc = time_function()       
    log.info('Packing finished %i iterations in %f seconds', i+1, toc-tic)
    log.info('%i/%i parts were packed successfully', np.sum(overall_inserted), len(polygons))
    log.info('Final rectangular density is %f.', overall_density)
    
    polygon_density = np.sum(polygon_area[overall_inserted])/np.product(overall_sheet)
    log.info('Final polygonal density is %f.', polygon_density)

    transforms_obb    = transforms_obb[overall_inserted]
    transforms_packed = transforms_obb.copy()
    transforms_packed.reshape(-1,9)[:,[2,5]] += overall_offset + buffer_dist
 
    if plot: 
        transform_polygon(np.array(polygons)[overall_inserted],  
                          transforms_packed, 
                          plot=True)
    rectangles -= 2.0*buffer_dist
    
    if return_all:
        return (overall_inserted, 
                transforms_packed, 
                transforms_obb, 
                overall_sheet,
                rectangles[overall_inserted])

    return overall_inserted, transforms_packed



'''
trimesh.py
========
Python library for loading triangular meshes and doing simple operations on them. Included loaders are binary/ASCII STL and Wavefront (OBJ), included exporters are binary STL or COLLADA. If Assimp/pyassimp are available, meshes can be loaded using the assimp loaders.

Using
-----
    >>> import trimesh
    >>> m = trimesh.load_mesh('models/ballA.off')
    >>> m.show()

'''
from .version     import __version__
from .base        import Trimesh
from .scene.scene import Scene

from .util    import unitize
from .transformations  import transform_points

from .io.load import load_mesh, load_path, load, available_formats

from . import transformations
from . import primitives
